from enum import Enum

import numpy as np
from numpy.typing import NDArray
from ophyd_async.core import ConfigSignal, DeviceVector, HintedSignal, StandardReadable
from ophyd_async.epics.signal import epics_signal_r, epics_signal_rw, epics_signal_x

from ..typing import StrEnum


def num_to_char(num):
    char = chr(65 + num)
    return char


class ScalerChannel(StandardReadable):
    def __init__(self, prefix, channel_num, name=""):
        epics_ch_num = channel_num + 1  # EPICS is 1-indexed
        # Hinted signals
        with self.add_children_as_readables(HintedSignal):
            net_suffix = (
                f"_net{num_to_char((channel_num // 12))}"
                f".{num_to_char(channel_num % 12)}"
            )
            self.net_count = epics_signal_r(float, f"{prefix}{net_suffix}")
        # Regular readable signals
        with self.add_children_as_readables():
            self.raw_count = epics_signal_r(float, f"{prefix}.S{epics_ch_num}")
        # Configuration signals
        with self.add_children_as_readables(ConfigSignal):
            self.description = epics_signal_rw(str, f"{prefix}.NM{epics_ch_num}")
            self.is_gate = epics_signal_rw(bool, f"{prefix}.G{epics_ch_num}")
            self.preset_count = epics_signal_rw(float, f"{prefix}.PR{epics_ch_num}")
            offset_suffix = f"_offset{channel_num // 4}.{num_to_char(channel_num % 4)}"
            self.offset_rate = epics_signal_rw(float, f"{prefix}{offset_suffix}")
        super().__init__(name=name)


class MCA(StandardReadable):

    class MCAMode(str, Enum):
        PHA = "PHA"
        MCS = "MCS"
        LIST = "List"

    def __init__(self, prefix, name=""):
        # Signals
        with self.add_children_as_readables(HintedSignal):
            self.spectrum = epics_signal_r(NDArray[np.int32], f"{prefix}.VAL")
        self.background = epics_signal_r(NDArray[np.int32], f"{prefix}.BG")
        with self.add_children_as_readables(ConfigSignal):
            self.mode = epics_signal_rw(self.MCAMode, f"{prefix}.MODE")
        super().__init__(name=name)


class MultiChannelScaler(StandardReadable):
    """A SIS3820 or MeasComp counter.

    Devices
    =======
    scaler
      The scaler-like device for counting channels once.
    mcas
      The multi-channel analyzers for measuring the channels
      repeatedly.

    """

    _ophyd_labels_ = {"scalers"}

    class ChannelAdvanceSource(str, Enum):
        INTERNAL = "Internal"
        EXTERNAL = "External"

    class Acquiring(str, Enum):
        DONE = "Done"
        ACQUIRING = "Acquiring"

    class ScalerModel(str, Enum):
        SIS_3801 = "SIS3801"
        SIS_3820 = "SIS3820"

    class Channel1Source(str, Enum):
        INTERNAL_CLOCK = "Int. clock"
        EXTERNAL = "External"

    class AcquireMode(str, Enum):
        MCS = "MCS"
        SCALER = "Scaler"

    class Polarity(str, Enum):
        NORMAL = "Normal"
        INVERTED = "Inverted"

    class OutputMode(str, Enum):
        MODE_0 = "Mode 0"
        MODE_1 = "Mode 1"
        MODE_2 = "Mode 2"
        MODE_3 = "Mode 3"

    class InputMode(str, Enum):
        MODE_0 = "Mode 0"
        MODE_1 = "Mode 1"
        MODE_2 = "Mode 2"
        MODE_3 = "Mode 3"
        MODE_4 = "Mode 4"
        MODE_5 = "Mode 5"
        MODE_6 = "Mode 6"

    def __init__(self, prefix, channels: list[int], name=""):
        # Controls
        self.start_all = epics_signal_x(f"{prefix}StartAll")
        self.stop_all = epics_signal_x(f"{prefix}StopAll")
        self.erase_all = epics_signal_x(f"{prefix}EraseAll")
        self.erase_start = epics_signal_x(f"{prefix}EraseStart")
        self.software_channel_advance = epics_signal_x(
            f"{prefix}SoftwareChannelAdvance"
        )
        # Transient states
        self.acquiring = epics_signal_r(self.Acquiring, f"{prefix}Acquiring")
        self.user_led = epics_signal_rw(bool, f"{prefix}UserLED")
        # Config signals
        with self.add_children_as_readables(ConfigSignal):
            self.preset_time = epics_signal_rw(float, f"{prefix}PresetReal")
            self.dwell_time = epics_signal_rw(float, f"{prefix}Dwell")
            self.prescale = epics_signal_rw(int, f"{prefix}Prescale")
            self.channel_advance_source = epics_signal_rw(
                self.ChannelAdvanceSource, f"{prefix}ChannelAdvance"
            )
            self.count_on_start = epics_signal_rw(bool, f"{prefix}CountOnStart")
            self.channel_1_source = epics_signal_rw(
                self.Channel1Source, f"{prefix}Channel1Source"
            )
            self.mux_output = epics_signal_rw(float, f"{prefix}MUXOutput")
            self.acquire_mode = epics_signal_rw(
                self.AcquireMode, f"{prefix}AcquireMode"
            )
            self.input_mode = epics_signal_rw(self.InputMode, f"{prefix}InputMode")
            self.input_polarity = epics_signal_rw(
                self.Polarity, f"{prefix}InputPolarity"
            )
            self.output_mode = epics_signal_rw(self.OutputMode, f"{prefix}OutputMode")
            self.output_polarity = epics_signal_rw(
                self.Polarity, f"{prefix}OutputPolarity"
            )
            self.lne_output_stretcher = epics_signal_rw(
                bool, f"{prefix}LNEStretcherEnable"
            )
            self.lne_output_polarity = epics_signal_rw(
                self.Polarity, f"{prefix}LNEOutputPolarity"
            )
            self.lne_output_delay = epics_signal_rw(float, f"{prefix}LNEOutputDelay")
            self.lne_output_width = epics_signal_rw(float, f"{prefix}LNEOutputWidth")
            self.num_channels_max = epics_signal_r(int, f"{prefix}MaxChannels")
            self.num_channels = epics_signal_rw(int, f"{prefix}NuseAll")
            self.snl_connected = epics_signal_r(bool, f"{prefix}SNL_Connected")
            self.model = epics_signal_r(self.ScalerModel, f"{prefix}Model")
            self.firmware = epics_signal_r(int, f"{prefix}Firmware")
        # Child-devices
        with self.add_children_as_readables():
            self.mcas = DeviceVector({i: MCA(f"{prefix}mca{i+1}") for i in channels})
            self.scaler = Scaler(f"{prefix}scaler1", channels=channels)
            self.elapsed_time = epics_signal_r(float, f"{prefix}ElapsedReal")
            self.current_channel = epics_signal_r(int, f"{prefix}CurrentChannel")
        super().__init__(name=name)


class Scaler(StandardReadable):
    """A scaler device that has one or more channels."""

    class CountMode(StrEnum):
        ONE_SHOT = "OneShot"
        AUTO_COUNT = "AutoCount"

    class CountState(StrEnum):
        DONE = "Done"
        COUNT = "Count"

    def __init__(self, prefix, channels: list[int], name=""):
        # Add invidiaul scaler channels
        with self.add_children_as_readables():
            # Add individual channels
            self.channels = DeviceVector(
                {
                    ch_num: ScalerChannel(f"{prefix}", channel_num=ch_num)
                    for ch_num in channels
                }
            )
        # Scaler-specific signals
        with self.add_children_as_readables():
            self.elapsed_time = epics_signal_r(float, f"{prefix}.T")
        with self.add_children_as_readables(ConfigSignal):
            self.delay = epics_signal_rw(float, f"{prefix}.DLY")
            self.clock_frequency = epics_signal_rw(float, f"{prefix}.FREQ")
            self.count_mode = epics_signal_rw(self.CountMode, f"{prefix}.CONT")
            self.preset_time = epics_signal_rw(float, f"{prefix}.TP")
        self.auto_count = epics_signal_rw(bool, f"{prefix}.CONT")
        self.count = epics_signal_rw(self.CountState, f"{prefix}.CNT")
        self.record_dark_current = epics_signal_x(f"{prefix}_offset_start.PROC")
        self.auto_count_delay = epics_signal_rw(float, f"{prefix}.DLY1")
        self.auto_count_time = epics_signal_rw(float, f"{prefix}.TP1")
        self.dark_current_time = epics_signal_rw(float, f"{prefix}_offset_time.VAL")
        super().__init__(name=name)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import os
import sys
from functools import partial

# This environmental variable needs to be set before importing haven
os.environ["HAVEN_CONFIG_FILES"] = f"{os.environ['BLUESKY_DIR']}/iconfig.toml"

import msgpack
import msgpack_numpy as mpn
from bluesky_kafka import MongoConsumer

import haven


def main():
    """Launch the mongo consumer."""
    bootstrap_servers = "localhost:9092"

    # Determine the mongo DB URI from the databroker/intake configuration
    catalog_name = haven.load_config()["database"]["databroker"]["catalog"]
    catalog = haven.load_catalog(name=catalog_name)
    host, port = catalog._resource_collection.database.client.address
    mongo_uri = f"mongodb://{host}:{port}"

    if mongo_uri is None:
        raise AttributeError("Environment variable BLUESKY_MONGO_URI " "must be set.")

    kafka_deserializer = partial(msgpack.loads, object_hook=mpn.decode)
    auto_offset_reset = "latest"
    topics = ["s25idc_queueserver", "s25idd_queueserver"]

    topic_database_map = {
        "s25idc_queueserver": "25idc-bluesky",
        "s25idd_queueserver": "25idd-bluesky",
    }

    # Create a MongoConsumer that will automatically listen to new beamline topics.
    # The parameter metadata.max.age.ms determines how often the consumer will check for
    # new topics. The default value is 5000ms.
    mongo_consumer = MongoConsumer(
        mongo_uri,
        topic_database_map,
        tls=False,
        topics=topics,
        bootstrap_servers=bootstrap_servers,
        group_id="mongodb",
        consumer_config={"auto.offset.reset": auto_offset_reset},
        polling_duration=1.0,
        deserializer=kafka_deserializer,
    )

    mongo_consumer.start()


if __name__ == "__main__":
    sys.exit(main())



from unittest import mock

import pytest
from ophyd_async.core import DEFAULT_TIMEOUT

from haven.devices.srs570 import GainSignal, SRS570PreAmplifier


@pytest.fixture()
async def preamp():
    preamp = SRS570PreAmplifier("255idcVEM:SR02:", name="")
    await preamp.connect(mock=True)
    return preamp


# Known settling times measured from the I0 SR-570 at 25-ID-C
settling_times = {
    # (sensitivity_value, sensitivity_unit, gain_mode): settle_time
    # pA/V
    ("1", "pA/V", "HIGH BW"): 2.5,  # 1 pA/V
    ("2", "pA/V", "HIGH BW"): 2.0,
    ("5", "pA/V", "HIGH BW"): 2.0,
    ("10", "pA/V", "HIGH BW"): 0.5,
    ("20", "pA/V", "HIGH BW"): 0.5,
    ("50", "pA/V", "HIGH BW"): 0.5,  # 50 pA/V
    ("100", "pA/V", "HIGH BW"): 0.5,
    ("200", "pA/V", "HIGH BW"): 0.3,
    ("500", "pA/V", "HIGH BW"): 0.3,
    # nA/V
    ("1", "nA/V", "HIGH BW"): 0.3,
    ("2", "nA/V", "HIGH BW"): 0.3,  # 2 nA/V
    ("5", "nA/V", "HIGH BW"): 0.3,
    ("10", "nA/V", "HIGH BW"): 0.3,
    ("20", "nA/V", "HIGH BW"): 0.3,
    ("50", "nA/V", "HIGH BW"): 0.3,
    ("100", "nA/V", "HIGH BW"): 0.3,  # 100 nA/V
    ("200", "nA/V", "HIGH BW"): 0.3,
    ("500", "nA/V", "HIGH BW"): 0.3,
    # μA/V
    ("1", "uA/V", "HIGH BW"): 0.3,
    ("2", "uA/V", "HIGH BW"): 0.3,
    ("5", "uA/V", "HIGH BW"): 0.3,  # 5 μA/V
    ("10", "uA/V", "HIGH BW"): 0.3,
    ("20", "uA/V", "HIGH BW"): 0.3,
    ("50", "uA/V", "HIGH BW"): 0.3,
    ("100", "uA/V", "HIGH BW"): 0.3,
    ("200", "uA/V", "HIGH BW"): 0.3,  # 200 μA/V
    ("500", "uA/V", "HIGH BW"): 0.3,
    # mA/V
    ("1", "mA/V", "HIGH BW"): 0.3,
    ("2", "mA/V", "HIGH BW"): None,
    ("5", "mA/V", "HIGH BW"): None,
    ("10", "mA/V", "HIGH BW"): None,
    ("20", "mA/V", "HIGH BW"): None,
    ("50", "mA/V", "HIGH BW"): None,
    ("100", "mA/V", "HIGH BW"): None,
    ("200", "mA/V", "HIGH BW"): None,
    ("500", "mA/V", "HIGH BW"): None,
    # pA/V
    ("1", "pA/V", "LOW NOISE"): 3.0,
    ("2", "pA/V", "LOW NOISE"): 2.5,
    ("5", "pA/V", "LOW NOISE"): 2.0,
    ("10", "pA/V", "LOW NOISE"): 2.0,
    ("20", "pA/V", "LOW NOISE"): 1.75,
    ("50", "pA/V", "LOW NOISE"): 1.5,
    ("100", "pA/V", "LOW NOISE"): 1.25,
    ("200", "pA/V", "LOW NOISE"): 0.5,
    ("500", "pA/V", "LOW NOISE"): 0.5,
    # nA/V
    ("1", "nA/V", "LOW NOISE"): 0.3,
    ("2", "nA/V", "LOW NOISE"): 0.3,
    ("5", "nA/V", "LOW NOISE"): 0.3,
    ("10", "nA/V", "LOW NOISE"): 0.3,
    ("20", "nA/V", "LOW NOISE"): 0.3,
    ("50", "nA/V", "LOW NOISE"): 0.3,
    ("100", "nA/V", "LOW NOISE"): 0.3,
    ("200", "nA/V", "LOW NOISE"): 0.3,
    ("500", "nA/V", "LOW NOISE"): 0.3,
    # μA/V
    ("1", "uA/V", "LOW NOISE"): 0.3,
    ("2", "uA/V", "LOW NOISE"): 0.3,
    ("5", "uA/V", "LOW NOISE"): 0.3,
    ("10", "uA/V", "LOW NOISE"): 0.3,
    ("20", "uA/V", "LOW NOISE"): 0.3,
    ("50", "uA/V", "LOW NOISE"): 0.3,
    ("100", "uA/V", "LOW NOISE"): 0.3,
    ("200", "uA/V", "LOW NOISE"): 0.3,
    ("500", "uA/V", "LOW NOISE"): 0.3,
    # mA/V
    ("1", "mA/V", "LOW NOISE"): 0.3,
    ("2", "mA/V", "LOW NOISE"): None,
    ("5", "mA/V", "LOW NOISE"): None,
    ("10", "mA/V", "LOW NOISE"): None,
    ("20", "mA/V", "LOW NOISE"): None,
    ("50", "mA/V", "LOW NOISE"): None,
    ("100", "mA/V", "LOW NOISE"): None,
    ("200", "mA/V", "LOW NOISE"): None,
    ("500", "mA/V", "LOW NOISE"): None,
}

gain_units = ["pA/V", "nA/V", "uA/V", "mA/V"]
gain_values = ["1", "2", "5", "10", "20", "50", "100", "200", "500"]
gain_modes = ["LOW NOISE", "HIGH BW"]


async def test_preamp_signals(preamp):
    # Check the enums
    await preamp.sensitivity_value.set(SRS570PreAmplifier.SensValue._5)


@pytest.mark.parametrize("gain_mode", gain_modes)
@pytest.mark.parametrize("gain_unit", gain_units)
@pytest.mark.parametrize("gain_value", gain_values)
async def test_preamp_gain_settling(gain_value, gain_unit, gain_mode, mocker, preamp):
    """The SR-570 Pre-amp voltage spikes when changing gain.

    One solution, tested here, is to add a dynamic settling time.

    """
    value_idx = gain_values.index(gain_value)
    unit_idx = gain_units.index(gain_unit)
    settle_time = settling_times[
        (
            gain_value,
            gain_unit,
            gain_mode,
        )
    ]
    # Create and check the device
    assert isinstance(preamp.sensitivity_unit, GainSignal)
    assert isinstance(preamp.sensitivity_value, GainSignal)
    # Set up a mocked sleep function
    sleep_mock = mock.AsyncMock()
    mocker.patch("haven.devices.ion_chamber.asyncio.sleep", sleep_mock)
    # Set the sensitivity based on value
    await preamp.sensitivity_unit.set(gain_unit)
    await preamp.gain_mode.set(gain_mode)
    sleep_mock.reset_mock()
    await preamp.sensitivity_value.set(gain_value)
    # Check that the signal's ``set`` was called with correct arguments
    preamp.sensitivity_value._backend.put_mock.assert_called_once_with(
        gain_value, wait=True, timeout=DEFAULT_TIMEOUT
    )
    # Check that the settle time was included
    sleep_mock.assert_called_once_with(settle_time)


async def test_preamp_gain_mode_settling(mocker, preamp):
    """The SR-570 Pre-amp also has a low drift mode, whose settling times
    are the same as the low noise mode.

    """
    gain_unit = "pA/V"
    gain_value = "500"
    settle_time = 0.5
    # Set up a mocked sleep function
    sleep_mock = mock.AsyncMock()
    mocker.patch("haven.devices.ion_chamber.asyncio.sleep", sleep_mock)
    # Set the preamp gains
    await preamp.sensitivity_unit.set(gain_unit)
    await preamp.sensitivity_value.set(gain_value)
    sleep_mock.reset_mock()
    await preamp.gain_mode.set("LOW DRIFT")
    # Check that the correct settle_time was used
    sleep_mock.assert_called_once_with(settle_time)


async def test_gain_signals(preamp):
    assert hasattr(preamp, "gain")
    assert hasattr(preamp, "gain_db")
    # Change the preamp settings
    await preamp.sensitivity_value.set("20")
    await preamp.sensitivity_unit.set("uA/V")
    await preamp.offset_value.set("2")
    await preamp.offset_unit.set("uA")
    # Check the gain and gain_db signals
    gain = await preamp.gain.get_value()
    assert gain == pytest.approx(1 / 20e-6)
    gain_db = await preamp.gain_db.get_value()
    assert gain_db == pytest.approx(46.9897)


@pytest.mark.skip(reason="Not sure put complete is needed with ophyd-async")
def test_sensitivity_put_complete():
    """Test the fix for a race condition in the pre-amp EPICS database.

    If the gain is set to 1 mA/V in the IOC, and you try to set it to
    e.g. 200 µA/V, you need to set the unit first then the value since
    "1" is the only valid "mA/V" value:

    ✓ 1 mA/V -> 1 µA/V -> 200 µA/V
    ✗ 1 mA/V -> 200 mA/V (corrects to 1 mA/V) -> 1 µA/V

    Even if the order of ``set`` calls is correct in ophyd, the
    requests may not make it through the calc/transform in the right
    order, so we wind up at the wrong gain.

    Apparently, using put_complete for the sensitivity unit fixes this
    problem, though it's not entirely clear why.

    Regardless, this test checks that put_complete is set on the given
    signal.

    """
    preamp = SRS570PreAmplifier("prefix:", name="preamp")
    put_complete_signals = [
        preamp.sensitivity_unit,
    ]
    for sig in put_complete_signals:
        assert sig._put_complete is True, sig


@pytest.mark.asyncio
async def test_get_gain_level(preamp):
    # Change the preamp settings
    await preamp.sensitivity_value.set("20")
    await preamp.sensitivity_unit.set("uA/V"),
    await preamp.offset_value.set("2"),  # 2 uA/V
    await preamp.offset_unit.set("uA/V"),
    # Check that the gain level moved
    gain_level = await preamp.gain_level.get_value()
    assert gain_level == 5


@pytest.mark.asyncio
async def test_put_gain_level(preamp):
    # Move the gain level
    await preamp.gain_level.set(15)
    # Check that the preamp sensitivities are moved
    assert await preamp.sensitivity_value.get_value() == "10"
    assert await preamp.sensitivity_unit.get_value() == "nA/V"
    # Check that the preamp sensitivity offsets are moved
    assert await preamp.offset_value.get_value() == "1"
    assert await preamp.offset_unit.get_value() == "nA"



import asyncio
import logging

from qasync import asyncSlot
from qtpy.QtCore import Signal

from firefly import display
from firefly.plans.util import is_valid_value, time_converter

log = logging.getLogger()


class RegionBase:
    def __init__(self, line_label: str = ""):
        self.line_label = line_label
        self.setup_ui()

    def setup_ui(self):
        raise NotImplementedError


class RegionsDisplay(display.FireflyDisplay):
    """Contains variable number of plan parameter regions in a table.

    Should be subclassed to produce a usable display.

    Expects the subclass to have the following attributes:

    ui.regions_layout
      The vertical layout to receive the paramter regions.
    ui.num_motor_spin_box
      A spin box widget for how many regions to show.
    ui.run_button
      A button widget that will queue the plan.
    Region
      The object representing the region view to put into the layout.
    queue_plan
      A method that emits to `queue_item_submitted` signal with a plan
      item.

    """

    default_num_regions = 1
    Region = RegionBase

    scan_time_changed = Signal(float)
    total_time_changed = Signal(float)

    def customize_ui(self):
        # Remove the default layout from .ui file
        try:
            self.clearLayout(self.ui.region_template_layout)
        except AttributeError:
            pass

        # Disable the line edits in spin box (use up/down buttons instead)
        self.ui.num_motor_spin_box.lineEdit().setReadOnly(True)

        # Create the initial (blank) regions
        self.regions = []
        self.ui.num_motor_spin_box.setValue(self.default_num_regions)
        self.add_regions(self.default_num_regions)
        # Set up the mechanism for changing region number
        self.ui.num_motor_spin_box.valueChanged.connect(self.update_regions_slot)
        self.ui.run_button.clicked.connect(self.queue_plan)

    @asyncSlot(object)
    async def update_devices(self, registry):
        """Set available components in the motor boxes."""
        await super().update_devices(registry)
        await self.update_component_selector_devices(registry)
        if hasattr(self.ui, "detectors_list"):
            await self.ui.detectors_list.update_devices(registry)

    async def update_component_selector_devices(self, registry):
        """Update the devices for all the component selectors"""
        selectors = [region.motor_box for region in self.regions]
        aws = [box.update_devices(registry) for box in selectors]
        await asyncio.gather(*aws)

    def clearLayout(self, layout):
        if layout is not None:
            while layout.count():
                item = layout.takeAt(0)
                if item.widget():
                    item.widget().deleteLater()

    def add_regions(self, num: int = 1):
        """Add *num* regions to the list of scan parameters.

        Returns
        =======
        new_regions
          The newly created region objects.

        """
        new_regions = []
        for i in range(num):
            region = self.Region(len(self.regions))
            self.ui.regions_layout.addLayout(region.layout)
            # Save it to the list
            self.regions.append(region)
            new_regions.append(region)
        # Finish setting up the new regions
        return new_regions

    def remove_regions(self, num=1):
        for i in range(num):
            layout = self.regions[-1].layout
            # iterate/wait, and delete all widgets in the layout in the end
            while layout.count() > 0:
                item = layout.takeAt(0)
                if item.widget():
                    item.widget().deleteLater()
            self.regions.pop()

    @asyncSlot(int)
    async def update_regions_slot(self, new_region_num):
        return await self.update_regions(new_region_num)

    async def update_regions(self, new_region_num: int):
        """Adjust regions from the scan params layout to reach
        *new_region_num*.

        """
        old_region_num = len(self.regions)
        diff_region_num = new_region_num - old_region_num
        new_regions = []
        if diff_region_num < 0:
            self.remove_regions(abs(diff_region_num))
        elif diff_region_num > 0:
            new_regions = self.add_regions(diff_region_num)
        # Setup the component selector devices with existing device definitions
        if len(new_regions) > 0:
            aws = [
                region.motor_box.update_devices(self.registry) for region in new_regions
            ]
            await asyncio.gather(*aws)

    async def _get_time(self, detector):
        """Get the dwell time value for a given detector."""
        time_signal = detector.default_time_signal
        if hasattr(time_signal, "get_value"):
            return await time_signal.get_value()
        return time_signal.get()

    @asyncSlot()
    async def update_total_time(self):
        # get default detector time
        detectors = self.ui.detectors_list.selected_detectors()
        detectors = [self.registry[name] for name in detectors]
        detectors = [det for det in detectors if hasattr(det, "default_time_signal")]

        # to prevent detector list is empty
        if len(detectors) == 0:
            detector_time = float("nan")
        else:
            detector_time = max([await self._get_time(det) for det in detectors])
        # get scan num points to calculate total time
        total_time_per_scan = self.time_per_scan(detector_time)

        # calculate time for each scan
        hrs, mins, secs = time_converter(total_time_per_scan)
        self.ui.label_hour_scan.setText(str(hrs))
        self.ui.label_min_scan.setText(str(mins))
        self.ui.label_sec_scan.setText(str(secs))

        self.scan_time_changed.emit(total_time_per_scan)

        # calculate time for entire plan
        num_scan_repeat = self.ui.spinBox_repeat_scan_num.value()
        total_time = num_scan_repeat * total_time_per_scan
        hrs_total, mins_total, secs_total = time_converter(total_time)

        self.ui.label_hour_total.setText(str(hrs_total))
        self.ui.label_min_total.setText(str(mins_total))
        self.ui.label_sec_total.setText(str(secs_total))
        self.total_time_changed.emit(total_time)

    def time_per_scan(self, detector_time):
        raise NotImplementedError

    def get_scan_parameters(self):
        # Get scan parameters from widgets
        detectors = self.ui.detectors_list.selected_detectors()
        repeat_scan_num = int(self.ui.spinBox_repeat_scan_num.value())

        # Get paramters from each rows of line regions:
        motor_lst, start_lst, stop_lst = [], [], []
        for region_i in self.regions:
            motor_lst.append(region_i.motor_box.current_component().name)
            start_lst.append(float(region_i.start_line_edit.text()))
            stop_lst.append(float(region_i.stop_line_edit.text()))

        motor_args = [
            values
            for motor_i in zip(motor_lst, start_lst, stop_lst)
            for values in motor_i
        ]

        return detectors, motor_args, repeat_scan_num

    def get_meta_data(self):
        # Get meta data info
        md = {
            "sample": self.ui.lineEdit_sample.text(),
            "purpose": self.ui.lineEdit_purpose.text(),
            "notes": self.ui.textEdit_notes.toPlainText(),
        }
        # Only include metadata that isn't an empty string
        md = {key: val for key, val in md.items() if is_valid_value(val)}
        return md



import uuid
from collections import OrderedDict, abc
from typing import Mapping, Sequence, Union

import numpy as np
from bluesky import plan_patterns
from bluesky import plan_stubs as bps
from bluesky import plans as bp
from bluesky.preprocessors import (
    __read_and_stash_a_motor,
    _normalize_devices,
    finalize_wrapper,
    pchain,
    plan_mutator,
    reset_positions_wrapper,
    run_wrapper,
    stage_wrapper,
)
from bluesky.protocols import EventPageCollectable
from bluesky.utils import Msg, single_gen
from ophyd import Device
from ophyd.flyers import FlyerInterface
from ophyd.status import StatusBase
from ophyd_async.core import TriggerInfo
from ophyd_async.epics.motor import FlyMotorInfo

from ..preprocessors import baseline_decorator

__all__ = ["fly_scan", "grid_fly_scan"]


def reset_flyers_wrapper(plan, devices=None):
    """Return flyer devices to their initial positions and velocities
    prior to being prepared for flying.

    Parameters
    ----------
    plan : iterable or iterator
        a generator, list, or similar containing `Msg` objects
    devices : collection or None, optional
        If default (None), apply to all devices that are prepared by the plan.

    Yields
    ------
    msg : Msg
        messages from plan with 'read' and finally 'set' messages inserted

    """
    initial_positions = OrderedDict()
    if devices is not None:
        devices, coupled_parents = _normalize_devices(devices)
    else:
        coupled_parents = set()

    def insert_reads(msg):
        eligible = devices is None or msg.obj in devices
        seen = msg.obj in initial_positions
        if (msg.command == "prepare") and eligible and not seen:
            # Stash flyer position
            plan_bits = [
                __read_and_stash_a_motor(msg.obj, initial_positions, coupled_parents)
            ]
            # Stash flyer velocity
            if hasattr(msg.obj, "velocity"):
                plan_bits.append(
                    __read_and_stash_a_motor(
                        msg.obj.velocity, initial_positions, coupled_parents
                    )
                )
            return (pchain(*plan_bits, single_gen(msg)), None)
        else:
            return None, None

    def reset():
        blk_grp = f"reset-{str(uuid.uuid4())[:6]}"
        for k, v in initial_positions.items():
            if k.parent in coupled_parents:
                continue
            yield Msg("set", k, v, group=blk_grp)
        yield Msg("wait", None, group=blk_grp)

    return (yield from finalize_wrapper(plan_mutator(plan, insert_reads), reset()))


def fly_line_scan(detectors: list, *args, num, dwell_time):
    """A plan stub for fly-scanning a single trajectory.

    Parameters
    ==========
    detectors
      Flyables that will be trigger before the movers move.
    *args
      For one dimension, motor, start, stop. In general:

      .. code-block:: python

         motor1, start1, stop1,
         motor2, start2, stop2,
         ...,
         motorN, startN, stopN

      Motors can be any ‘flyable’ object.
    num
      Number of measurements to take.
    dwell_time
      How long, in seconds, for each measurement point.
    """
    # Calculate parameters for the fly-scan
    # step_size = abs(start - stop) / (num - 1)
    motors = args[0::3]
    starts = args[1::3]
    ends = args[2::3]
    # Set up motors in their taxi position
    prepare_group = uuid.uuid4()
    for obj, start, end in zip(motors, starts, ends):
        position_info = FlyMotorInfo(
            start_position=start,
            end_position=end,
            time_for_move=dwell_time * num,
        )
        yield from bps.prepare(obj, position_info, wait=False, group=prepare_group)
    # Set up detectors
    trigger_info = TriggerInfo(
        number=num, livetime=dwell_time, deadtime=0, trigger="internal"
    )
    for obj in detectors:
        yield from bps.prepare(obj, trigger_info, wait=False, group=prepare_group)
    yield from bps.wait(group=prepare_group)
    # Monitor the motors during their move
    for motor in motors:
        sig = motor.user_readback
        yield from bps.monitor(sig, name=sig.name)
    # Perform the fly scan
    flyers = [*motors, *detectors]
    kickoff_group = uuid.uuid4()
    for flyer in flyers:
        yield from bps.kickoff(flyer, wait=False, group=kickoff_group)
    yield from bps.wait(group=kickoff_group)
    # Wait for all the flyers to be done
    motor_complete_group = uuid.uuid4()
    for m in motors:
        yield from bps.complete(m, wait=False, group=motor_complete_group)
    yield from bps.wait(group=(motor_complete_group))
    # Stop detectors
    det_complete_group = uuid.uuid4()
    for det in detectors:
        yield from bps.complete(det, wait=False, group=det_complete_group)
    yield from bps.wait(group=(det_complete_group))
    # Stop monitoring motors
    for motor in motors:
        sig = motor.user_readback
        yield from bps.unmonitor(sig)
    # Collect the data after flying
    flyers = [*motors, *detectors]
    flyers = [flyer for flyer in flyers if isinstance(flyer, EventPageCollectable)]
    for flyer_ in flyers:
        print(f"Collecting from {flyer_}")
        yield from bps.collect(flyer_)


@baseline_decorator()
def fly_scan(
    detectors: Sequence[FlyerInterface],
    *args,
    num: int,
    dwell_time: float,
    md: Mapping = {},
):
    """Do a fly scan with a 'flyer' motor and some 'flyer' detectors.

    Parameters
    ----------
    detectors
      List of 'readable' objects that support the flyer interface
    *args
      For one dimension, motor, start, stop. In general:

      .. code-block:: python

         motor1, start1, stop1,
         motor2, start2, stop2,
         ...,
         motorN, startN, stopN

      Motors can be any ‘flyable’ object.
    num
      Number of measurements to take.
    dwell_time
      How long, in seconds, for each measurement point.
    md
      metadata

    Yields
    ------
    msg
      'prepare', 'kickoff', 'complete, and 'collect' messages

    """
    # Stage the devices
    motors = args[0::3]
    starts = args[1::3]
    stops = args[2::3]
    devices = [*motors, *detectors]
    # Prepare metadata representation of the motor arguments
    md_args = zip([repr(m) for m in motors], starts, stops)
    md_args = tuple(obj for m, start, stop in md_args for obj in [m, start, stop])
    # Prepare metadata
    md_ = {
        "plan_name": "fly_scan",
        "motors": [motor.name for motor in motors],
        "detectors": [det.name for det in detectors],
        "plan_args": {
            "detectors": list(map(repr, detectors)),
            "*args": md_args,
            "num": num,
            "dwell_time": dwell_time,
        },
    }
    md_.update(md)
    # Execute the plan
    line_scan = fly_line_scan(
        detectors,
        *args,
        num=num,
        dwell_time=dwell_time,
    )
    # Wrapper for reseting the initial motor position/velocity
    line_scan = reset_flyers_wrapper(line_scan, motors)
    # Wrapper for making it a proper run
    line_scan = run_wrapper(line_scan, md=md_)
    line_scan = stage_wrapper(line_scan, motors)
    yield from line_scan


# @baseline_decorator()
def grid_fly_scan(
    detectors: Sequence[FlyerInterface],
    *args,
    snake_axes: Union[bool, Sequence[Device]] = False,
    md: Mapping = {},
):
    """Scan over a mesh with one of the axes collecting without stopping.

    Parameters
    ----------
    detectors
      list of 'readable' objects
    *args
      patterned like::

        motor1, start1, stop1, num1,
        motor2, start2, stop2, num2,
        ...
        flyer, flyer_start, flyer_stop, flyer_num

      The first motor is the "slowest", the outer loop. The last
      motor should be flyable.
    snake_axes
      which axes should be snaked, either ``False`` (do not snake any axes),
      ``True`` (snake all axes) or a list of axes to snake. "Snaking" an axis
      is defined as following snake-like, winding trajectory instead of a
      simple left-to-right trajectory. The elements of the list are motors
      that are listed in `args`. The list must not contain the slowest
      (first) motor, since it can't be snaked.
    md: dict, optional
      metadata

    Yields
    ------
    msg
      'stage', 'open_run', 'mv', 'kickoff', 'wait', 'complete, 'wait',
      'collect', 'close_run', 'stage' messages.

    """
    # Extract the step-scan vs fly-scan arguments
    *step_args, flyer, fly_start, fly_stop, fly_num = args
    # Handle giving snaked axes as a list
    step_chunks = list(plan_patterns.chunk_outer_product_args(step_args))
    num_steppers = len(step_chunks)
    motors = [m[0] for m in step_chunks]
    all_motors = [*motors, flyer]
    if isinstance(snake_axes, abc.Iterable) and not isinstance(snake_axes, str):
        snake_steppers = snake_axes.copy()
        try:
            snake_steppers.remove(flyer)
        except ValueError:
            snake_flyer = False
        else:
            snake_flyer = True
        # Save for metadata processing
        snaking = [
            (motor in snake_steppers) for motor, start, stop, num, snake in step_chunks
        ]
        snaking = (False, *snaking[1:], snake_flyer)
    else:
        snake_steppers = snake_axes
        snake_flyer = snake_axes
        snaking = [False, *[snake_axes for _ in step_chunks[1:]], snake_flyer]
    # Prepare metadata
    chunk_args = list(plan_patterns.chunk_outer_product_args(args))
    md_args = []
    motor_names = []
    for i, (motor, start, stop, num, snake) in enumerate(chunk_args):
        md_args.extend([repr(motor), start, stop, num])
        motor_names.append(motor.name)
    num_points = np.prod([num for motor, start, stop, num, snake in chunk_args])
    md_ = {
        "shape": tuple(num for motor, start, stop, num, snake in chunk_args),
        "extents": tuple(
            [start, stop] for motor, start, stop, num, snake in chunk_args
        ),
        "plan_args": {
            "detectors": list(map(repr, detectors)),
            "args": md_args,
        },
        "plan_name": "grid_fly_scan",
        "num_points": num_points,
        "num_intervals": num_points - 1,
        "motors": tuple(motor_names),
        "snaking": snaking,
        "hints": {},
    }
    # Add metadata hints for plotting, etc
    md_["hints"].setdefault("gridding", "rectilinear")
    try:
        md_["hints"].setdefault(
            "dimensions", [(m.hints["fields"], "primary") for m in all_motors]
        )
    except (AttributeError, KeyError):
        ...
    md_.update(md)
    # Set up the plan
    per_step = Snaker(
        snake_axes=snake_flyer,
        flyer=flyer,
        start=fly_start,
        stop=fly_stop,
        num=fly_num,
        extra_signals=motors,
    )
    grid_scan = yield from bp.grid_scan(
        detectors,
        *step_args,
        snake_axes=snake_steppers,
        per_step=per_step,
        md=md_,
    )
    grid_scan = reset_flyers_wrapper(grid_scan)
    grid_scan = reset_positions_wrapper(grid_scan)
    uid = yield from grid_scan
    return uid


class Snaker:
    """Executes 1-D fly line scans in a snaking motion.

    Each call of the snaker executes the same fly-scan, but supports
    alternating scan directions to save time.

    This callable's signature is compatible with bluesky's
    ``per_step`` option for scans, so instead of just reading a
    detector, a whole fly-scan is performed."

    """

    reverse: bool = False

    def __init__(self, snake_axes, flyer, start, stop, num, extra_signals):
        self.snake_axes = snake_axes
        self.flyer = flyer
        self.start = start
        self.stop = stop
        self.num = num
        self.extra_signals = extra_signals

    def __call__(self, detectors, step, pos_cache):
        # Move the step-scanning motors to the correct position
        yield from bps.move_per_step(step, pos_cache)
        # Determine line scans range based on snaking
        start, stop = (self.start, self.stop)
        if self.reverse and self.snake_axes:
            start, stop = stop, start
        self.reverse = not self.reverse
        # Launch the fly scan
        yield from fly_line_scan(
            detectors,
            self.flyer,
            start,
            stop,
            num=self.num,
            extra_signals=step.keys(),
        )


class FlyerCollector(FlyerInterface, Device):
    stream_name: str
    detectors: Sequence
    positioners: Sequence

    def __init__(
        self,
        detectors,
        positioners,
        stream_name: str = "primary",
        extra_signals=(),
        *args,
        **kwargs,
    ):
        # self.flyers = flyers
        self.detectors = detectors
        self.positioners = positioners
        self.stream_name = stream_name
        self.extra_signals = extra_signals
        super().__init__(*args, **kwargs)

    def kickoff(self):
        return StatusBase(success=True)

    def complete(self):
        return StatusBase(success=True)

    def collect(self):
        collections = [iter(flyer.collect()) for flyer in self.detectors]
        while True:
            event = {
                "data": {},
                "timestamps": {},
            }
            try:
                for coll in collections:
                    datum = next(coll)
                    event["data"].update(datum["data"])
                    event["timestamps"].update(datum["timestamps"])
            except StopIteration:
                break
            # Use the median time stamps for the overall event time
            timestamps = []
            for ts in event["timestamps"].values():
                timestamps.extend(np.asarray(ts).flatten())
            event["time"] = np.median(timestamps)
            # Add interpolated motor positions
            for motor in self.positioners:
                datum = motor.predict(event["time"])
                event["data"].update(datum["data"])
                event["timestamps"].update(datum["timestamps"])
            # Add extra non-flying signals (not inc. in event time)
            for signal in self.extra_signals:
                for signal_name, reading in signal.read().items():
                    event["data"][signal_name] = reading["value"]
                    event["timestamps"][signal_name] = reading["timestamp"]
            yield event

    def describe_collect(self):
        desc = OrderedDict()
        for flyer in [*self.positioners, *self.detectors]:
            for stream, this_desc in flyer.describe_collect().items():
                desc.update(this_desc)
        # Add extra signals, e.g. slow motor during a grid fly scan
        for signal in self.extra_signals:
            desc.update(signal.describe())
        return {self.stream_name: desc}


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from apstools.synApps import TransformRecord
from ophyd_async.core import Device

from .motor import Motor
from .transform import TransformRecord


class HighHeatLoadMirror(Device):
    """A single mirror, controlled by several motors.

    Possibly also bendable.
    """

    _ophyd_labels_ = {"mirrors"}

    def __init__(self, prefix: str, name: str = "", bendable=False):
        # Physical motors
        self.transverse = Motor(f"{prefix}m1")
        self.roll = Motor(f"{prefix}m2")
        self.upstream = Motor(f"{prefix}m3")
        self.downstream = Motor(f"{prefix}m4")

        # Pseudo motors
        self.pitch = Motor(f"{prefix}coarsePitch")
        self.normal = Motor(f"{prefix}lateral")

        # Standard transform records for the pseudo motors
        self.drive_transform = TransformRecord(f"{prefix}lats:Drive")
        self.readback_transform = TransformRecord(f"{prefix}lats:Readback")

        if bendable:
            self.bender = Motor(f"{prefix}m5")

        super().__init__(name=name)


class KBMirror(Device):
    """A single mirror in a KB mirror set."""

    def __init__(
        self,
        prefix: str,
        upstream_motor: str,
        downstream_motor: str,
        upstream_bender: str = "",
        downstream_bender: str = "",
        name: str = "",
    ):
        self.pitch = Motor(f"{prefix}pitch")
        self.normal = Motor(f"{prefix}height")
        self.upstream = Motor(upstream_motor)
        self.downstream = Motor(downstream_motor)
        if upstream_bender != "":
            self.bender_upstream = Motor(upstream_bender)
        if downstream_bender != "":
            self.bender_downstream = Motor(downstream_bender)
        # The pseudo motor transform records have
        # a missing ':', so we need to remove it.
        transform_prefix = "".join(prefix.rsplit(":", 2))
        self.drive_transform = TransformRecord(f"{transform_prefix}:Drive")
        self.readback_transform = TransformRecord(f"{transform_prefix}:Readback")


class KBMirrors(Device):
    _ophyd_labels_ = {"kb_mirrors"}

    def __init__(
        self,
        prefix: str,
        horiz_upstream_motor: str,
        horiz_downstream_motor: str,
        vert_upstream_motor: str,
        vert_downstream_motor: str,
        horiz_upstream_bender: str = "",
        horiz_downstream_bender: str = "",
        vert_upstream_bender: str = "",
        vert_downstream_bender: str = "",
        name: str = "",
    ):
        # Create the two sub-mirrors
        self.horiz = KBMirror(
            prefix=f"{prefix}H:",
            upstream_motor=horiz_upstream_motor,
            downstream_motor=horiz_downstream_motor,
            upstream_bender=horiz_upstream_bender,
            downstream_bender=horiz_downstream_bender,
        )
        self.vert = KBMirror(
            prefix=f"{prefix}V:",
            upstream_motor=vert_upstream_motor,
            downstream_motor=vert_downstream_motor,
            upstream_bender=vert_upstream_bender,
            downstream_bender=vert_downstream_bender,
        )
        super().__init__(name=name)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest

from firefly.mirror import MirrorDisplay
from haven.devices.mirrors import HighHeatLoadMirror


@pytest.fixture()
async def hhl_bendable_mirror(sim_registry):
    """A fake set of slits using the 4-blade setup."""
    mirr = HighHeatLoadMirror(prefix="255ida:ORM1:", name="hhl_mirror", bendable=True)
    await mirr.connect(mock=True)
    sim_registry.register(mirr)
    return mirr


async def test_bendable_mirror(hhl_bendable_mirror, qtbot):
    mirror = hhl_bendable_mirror
    display = MirrorDisplay(macros={"DEVICE": mirror.name})
    qtbot.addWidget(display)
    # Check that the bender controls are unlocked
    assert display.ui.bender_embedded_display.isEnabled()



"""A bluesky plan to scan the X-ray beam slits and then move the slits
to that position.

"""

import logging
import warnings

import lmfit
import numpy as np
from bluesky import plan_stubs as bps
from bluesky import plans as bp
from bluesky.callbacks import LiveFit
from bluesky.preprocessors import subs_decorator

__all__ = ["align_slits"]


log = logging.getLogger(__name__)


def align_slit_scan(slit_motor, ion_chamber):
    yield from bp.rel_scan([ion_chamber], slit_motor, -1, 1, 50)


def align_slits(slit_motors=[], ion_chamber=None):
    """Scan a range of values for the slit motor, and position the
    motor in the center.

    """
    for motor in slit_motors:
        # Set up the live fit object to process the results
        def gaussian(x, A, sigma, x0):
            return A * np.exp(-((x - x0) ** 2) / (2 * sigma**2))

        model = lmfit.Model(gaussian)
        init_guess = {"A": 2, "sigma": lmfit.Parameter("sigma", 3, min=0), "x0": -0.2}
        fit = LiveFit(model, ion_chamber.name, {"x": motor.name}, init_guess)
        scan = subs_decorator(fit)(align_slit_scan)
        # Execute the scan to find the center
        yield from scan(ion_chamber=ion_chamber, slit_motor=motor)
        # Move the slit motor to the center of the Gaussian peak
        new_center = fit.result.values["x0"]
        yield from bps.mv(motor, new_center)
        fit_quality = fit.result.chisqr
        # Logging, severity determined by quality of fit
        log.info(f"Centered slit '{motor.name}': {new_center} (X²: {fit_quality:.2f})")
        if fit_quality > 1:
            msg = (
                f"Poor fit while centering motor '{motor.name}'. X²: {fit_quality:.2f}"
            )
            log.warning(msg)
            warnings.warn(msg, RuntimeWarning)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import datetime as dt
import logging
from collections import OrderedDict
from typing import Mapping, Sequence

import numpy as np
import pandas as pd
from tiled import queries

from haven import exceptions
from haven.catalog import Catalog

log = logging.getLogger(__name__)


class DatabaseWorker:
    selected_runs: Sequence = []

    def __init__(self, catalog=None, *args, **kwargs):
        if catalog is None:
            catalog = Catalog()
        self.catalog = catalog
        super().__init__(*args, **kwargs)

    async def filtered_nodes(self, filters: Mapping):
        case_sensitive = False
        log.debug(f"Filtering nodes: {filters}")
        filter_params = [
            # (filter_name, query type, metadata key)
            ("user", queries.Regex, "proposal_users"),
            ("proposal", queries.Regex, "proposal_id"),
            ("esaf", queries.Regex, "esaf_id"),
            ("sample", queries.Regex, "sample_name"),
            # ('exit_status', queries.Regex, "exit_status"),
            ("plan", queries.Regex, "plan_name"),
            ("edge", queries.Regex, "edge"),
        ]
        # Apply filters
        runs = self.catalog
        for filter_name, Query, md_name in filter_params:
            val = filters.get(filter_name, "")
            if val != "":
                runs = await runs.search(
                    Query(md_name, val, case_sensitive=case_sensitive)
                )
        full_text = filters.get("full_text", "")
        if full_text != "":
            runs = await runs.search(
                queries.FullText(full_text, case_sensitive=case_sensitive)
            )
        return runs

    async def load_distinct_fields(self):
        """Get distinct metadata fields for filterable metadata.

        Emits
        =====
        distinct_fields_changed
          Emitted with the new dictionary of distinct metadata choices
          for each metadata key.

        """
        new_fields = {}
        target_fields = [
            "sample_name",
            "proposal_users",
            "proposal_id",
            "esaf_id",
            "sample_name",
            "plan_name",
            "edge",
        ]
        # Get fields from the database
        response = await self.catalog.distinct(*target_fields)
        # Build into a new dictionary
        for key, result in response["metadata"].items():
            field = key.split(".")[-1]
            new_fields[field] = [r["value"] for r in result]
        return new_fields

    async def load_all_runs(self, filters: Mapping = {}):
        all_runs = []
        nodes = await self.filtered_nodes(filters=filters)
        async for uid, node in nodes.items():
            # Get meta-data documents
            metadata = await node.metadata
            start_doc = metadata.get("start")
            if start_doc is None:
                log.debug(f"Skipping run with no start doc: {uid}")
                continue
            stop_doc = metadata.get("stop")
            if stop_doc is None:
                stop_doc = {}
            # Get a human-readable timestamp for the run
            timestamp = start_doc.get("time")
            if timestamp is None:
                run_datetime = ""
            else:
                run_datetime = dt.datetime.fromtimestamp(timestamp)
                run_datetime = run_datetime.strftime("%Y-%m-%d %H:%M:%S")
            # Get the X-ray edge scanned
            edge = start_doc.get("edge")
            E0 = start_doc.get("E0")
            E0_str = "" if E0 is None else str(E0)
            if edge and E0:
                edge_str = f"{edge} ({E0} eV)"
            elif edge:
                edge_str = edge
            elif E0:
                edge_str = E0_str
            else:
                edge_str = ""
            # Build the table item
            # Get sample data from: dd80f432-c849-4749-a8f3-bdeec6f9c1f0
            run_data = OrderedDict(
                plan_name=start_doc.get("plan_name", ""),
                sample_name=start_doc.get("sample_name", ""),
                edge=edge_str,
                E0=E0_str,
                exit_status=stop_doc.get("exit_status", ""),
                run_datetime=run_datetime,
                uid=uid,
                proposal_id=start_doc.get("proposal_id", ""),
                esaf_id=start_doc.get("esaf_id", ""),
                esaf_users=start_doc.get("esaf_users", ""),
            )
            all_runs.append(run_data)
        return all_runs

    async def signal_names(self, hinted_only: bool = False):
        """Get a list of valid signal names (data columns) for selected runs.

        Parameters
        ==========
        hinted_only
          If true, only signals with the kind="hinted" parameter get
          picked.

        """
        xsignals, ysignals = [], []
        for run in self.selected_runs:
            if hinted_only:
                xsig, ysig = await run.hints()
            else:
                df = await run.to_dataframe()
                xsig = ysig = df.columns
            xsignals.extend(xsig)
            ysignals.extend(ysig)
        # Remove duplicates
        xsignals = list(dict.fromkeys(xsignals))
        ysignals = list(dict.fromkeys(ysignals))
        return list(xsignals), list(ysignals)

    async def metadata(self):
        """Get all metadata for the selected runs in one big dictionary."""
        md = {}
        for run in self.selected_runs:
            md[run.uid] = await run.metadata
        return md

    async def load_selected_runs(self, uids):
        # Prepare the query for finding the runs
        uids = list(dict.fromkeys(uids))
        # Retrieve runs from the database
        runs = [await self.catalog[uid] for uid in uids]
        # runs = await asyncio.gather(*run_coros)
        self.selected_runs = runs
        return runs

    async def images(self, signal):
        """Load the selected runs as 2D or 3D images suitable for plotting."""
        images = OrderedDict()
        for idx, run in enumerate(self.selected_runs):
            # Load datasets from the database
            try:
                image = await run[signal]
            except KeyError:
                log.warning(f"Signal {signal} not found in run {run}.")
            else:
                images[run.uid] = image
        return images

    async def all_signals(self, hinted_only=False):
        """Produce dataframe with all signals for each run.

        The keys of the dictionary are the labels for each curve, and
        the corresponding value is a pandas dataframe with the scan data.

        """
        xsignals, ysignals = await self.signal_names(hinted_only=hinted_only)
        # Build the dataframes
        dfs = OrderedDict()
        for run in self.selected_runs:
            # Get data from the database
            df = await run.to_dataframe(signals=xsignals + ysignals)
            dfs[run.uid] = df
        return dfs

    async def signals(
        self,
        x_signal,
        y_signal,
        r_signal=None,
        use_log=False,
        use_invert=False,
        use_grad=False,
    ) -> Mapping:
        """Produce a dictionary with the 1D datasets for plotting.

        The keys of the dictionary are the labels for each curve, and
        the corresponding value is a pandas dataset with the data for
        each signal.

        """
        # Check for sensible inputs
        use_reference = r_signal is not None
        if "" in [x_signal, y_signal] or (use_reference and r_signal == ""):
            msg = (
                f"Empty signal name requested: x={repr(x_signal)}, y={repr(y_signal)},"
                f" r={repr(r_signal)}"
            )
            log.debug(msg)
            raise exceptions.EmptySignalName(msg)
        signals = [x_signal, y_signal]
        if use_reference:
            signals.append(r_signal)
        # Remove duplicates
        signals = list(dict.fromkeys(signals).keys())
        # Build the dataframes
        dfs = OrderedDict()
        for run in self.selected_runs:
            # Get data from the database
            df = await run.to_dataframe(signals=signals)
            # Check for missing signals
            missing_x = x_signal not in df.columns
            missing_y = y_signal not in df.columns
            missing_r = r_signal not in df.columns
            if missing_x or missing_y or (use_reference and missing_r):
                log.warning(
                    "Could not find signals {x_signal}, {y_signal} and {r_signal}"
                )
                continue
            # Apply transformations
            if use_reference:
                df[y_signal] /= df[r_signal]
            if use_log:
                df[y_signal] = np.log(df[y_signal])
            if use_invert:
                df[y_signal] *= -1
            if use_grad:
                df[y_signal] = np.gradient(df[y_signal], df[x_signal])
            series = pd.Series(df[y_signal].values, index=df[x_signal].values)
            dfs[run.uid] = series
        return dfs

    async def export_runs(self, filenames: Sequence[str], formats: Sequence[str]):
        for filename, run, format in zip(filenames, self.selected_runs, formats):
            await run.export(filename, format=format)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



#!/usr/bin/env python3
from caproto.server import (
    PVGroup,
    PvpropertyInteger,
    SubGroup,
    ioc_arg_parser,
    pvproperty,
    run,
)


class AreaDetectorGroup(PVGroup):
    """An IOC matching an EPICS area detector.

    Currently does not describe a full area detector setup. It is just
    meant to provide the bare essentials.

    """

    class CameraGroup(PVGroup):
        acquire_rbv = pvproperty(value=0, name="Acquire_RBV", dtype=PvpropertyInteger)
        acquire = pvproperty(value=0, name="Acquire", dtype=PvpropertyInteger)
        acquire_busy = pvproperty(value=0, name="AcquireBusy", dtype=PvpropertyInteger)
        gain = pvproperty(value=10, name="Gain")
        gain_rbv = pvproperty(value=10, name="Gain_RBV")

    cam = SubGroup(CameraGroup, prefix="cam1:")


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="255idSimDet:", desc="haven.tests.ioc_area_detector test IOC"
    )
    ioc = AreaDetectorGroup(**ioc_options)
    run(ioc.pvdb, **run_options)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
from functools import partial
from typing import Sequence

import qtawesome as qta
from bluesky_queueserver_api import BPlan
from pydm.widgets import PyDMChannel, PyDMLabel, PyDMPushButton
from pydm.widgets.analog_indicator import PyDMAnalogIndicator
from pydm.widgets.display_format import DisplayFormat
from qasync import asyncSlot
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import (
    QCheckBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
)

import haven
from firefly import display

log = logging.getLogger(__name__)


class VoltmetersDisplay(display.FireflyDisplay):
    _ion_chamber_rows: Sequence

    # Signals
    details_window_requested = Signal(str)  # ion-chamber device name

    def customize_ui(self):
        # Connect support for running the auto_gain and dark current plans
        self.ui.auto_gain_button.setToolTip(haven.auto_gain.__doc__)
        self.ui.auto_gain_button.clicked.connect(self.run_auto_gain)
        self.ui.dark_current_button.clicked.connect(self.record_dark_current)
        # Adjust layouts
        self.ui.voltmeters_layout.setHorizontalSpacing(0)
        self.ui.voltmeters_layout.setVerticalSpacing(0)

    def clear_layout(self, layout):
        if layout is None:
            return
        while layout.count() > 0:
            item = layout.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.deleteLater()
            else:
                self.clear_layout(item.layout())

    @asyncSlot(object)
    async def update_devices(self, registry):
        ion_chambers = registry.findall(label="ion_chambers")
        self.ion_chambers = sorted(
            ion_chambers, key=lambda c: c.scaler_channel.raw_count.source
        )
        # Clear the voltmeters grid layout
        self.clear_layout(self.voltmeters_layout)
        # Add embedded displays for all the ion chambers
        self._ion_chamber_rows = []
        for row_idx, ic in enumerate(self.ion_chambers):
            # Create the display object
            row = Row(number=row_idx, ion_chamber=ic)
            self._ion_chamber_rows.append(row)
            # Add widgets to the grid layout
            for col_idx, layout in enumerate(row.column_layouts):
                self.ui.voltmeters_layout.addLayout(layout, row_idx, col_idx)
            # Connect the details button signal
            details_slot = partial(self.details_window_requested.emit, ic.name)
            row.details_button.clicked.connect(details_slot)

    def update_queue_status(self, status):
        super().update_queue_status(status)
        # Update widgets when the queue status changes
        self.ui.auto_gain_button.update_queue_style(status)
        self.ui.dark_current_button.update_queue_style(status)

    def run_auto_gain(self):
        """Send a plan to the queueserver to auto-gain the pre-amps."""
        # Get plan arguments from display widgets
        kw = {}
        volts_min = self.ui.volts_min_line_edit.text()
        if volts_min != "":
            kw["volts_min"] = float(volts_min)
        volts_max = self.ui.volts_max_line_edit.text()
        if volts_max != "":
            kw["volts_max"] = float(volts_max)
        # Check which ion chambers to run the plan with
        ic_names = []
        for row in self._ion_chamber_rows:
            if row.auto_gain_checkbox.isChecked():
                ic_names.append(row.device.name)
        # Construct the plan
        item = BPlan("auto_gain", ic_names, **kw)
        # Send it to the queue server
        self.queue_item_submitted.emit(item)

    def record_dark_current(self):
        """Add an item to queueserver to record the dark current of the ion
        chambers.

        """
        # Determine which shutters to close
        shutters = []
        if self.ui.shutter_checkbox.isChecked():
            shutters.append("experiment_shutter")
        # Construct the plan
        ic_names = [ic.name for ic in self.ion_chambers]
        item = BPlan("record_dark_current", ic_names, shutters=shutters)
        # Send it to the queue server
        self.queue_item_submitted.emit(item)

    def ui_filename(self):
        return "voltmeters.ui"


class Row:
    """An row in the voltmeters display for a single ion chamber's signal."""

    def __init__(self, parent=None, *args, number: int, ion_chamber):
        self.parent = parent
        self.number = number
        self.device = ion_chamber
        self.setup_ui()

    def setup_ui(self):
        # Create container layouts for each column
        num_columns = 5
        self.column_layouts = []
        for idx in range(num_columns):
            layout = QVBoxLayout()
            self.column_layouts.append(layout)
        ##################
        # Create widgets #
        ##################
        device_name = self.device.name
        # Description label
        self.name_label = PyDMLabel(
            parent=self.parent,
            init_channel=f"haven://{device_name}.scaler_channel.description",
        )
        self.name_label.setStyleSheet('font: 12pt "Sans Serif";\nfont-weight: bold;')
        self.column_layouts[0].addWidget(self.name_label)
        # Analog indicator
        self.voltage_indicator = PyDMAnalogIndicator(
            parent=self.parent,
            init_channel=f"haven://{device_name}.voltmeter_channel.final_value",
        )
        self.voltage_indicator.showValue = False
        self.voltage_indicator.limitsFromChannel = False
        self.voltage_indicator.minorAlarmFromChannel = False
        self.voltage_indicator.majorAlarmFromChannel = False
        self.voltage_indicator.userLowerLimit = 0.0
        self.voltage_indicator.userUpperLimit = 5.0
        self.voltage_indicator.userUpperMinorAlarm = 4.5
        self.voltage_indicator.userLowerMinorAlarm = 0.5
        self.voltage_indicator.userUpperMajorAlarm = 5.0
        self.voltage_indicator.userLowerMajorAlarm = 0.15
        self.column_layouts[1].addWidget(self.voltage_indicator)
        # Voltage labels
        self.column_layouts[2].addItem(VSpacer())
        self.voltage_label_layout = QHBoxLayout()
        self.voltage_label_layout.setSpacing(3)
        self.column_layouts[2].addLayout(self.voltage_label_layout)
        self.voltage_label_layout.addItem(HSpacer())
        self.voltage_label = PyDMLabel(
            parent=self.parent,
            init_channel=f"haven://{device_name}.voltmeter_channel.final_value",
        )
        self.voltage_label.setStyleSheet('font: 12pt "Sans Serif";\nfont-weight: bold;')
        self.voltage_label_layout.addWidget(self.voltage_label)
        self.voltage_unit_label = QLabel(parent=self.parent)
        self.voltage_unit_label.setStyleSheet(
            'font: 12pt "Sans Serif";\nfont-weight: bold;'
        )
        self.voltage_unit_label.setText("V")
        self.voltage_label_layout.addWidget(self.voltage_unit_label)
        self.voltage_label_layout.addItem(HSpacer())
        # Current labels
        self.current_label_layout = QHBoxLayout()
        self.current_label_layout.setSpacing(3)
        self.column_layouts[2].addLayout(self.current_label_layout)
        self.current_label_layout.addItem(HSpacer())
        self.current_label = PyDMLabel(
            parent=self.parent,
            init_channel=f"haven://{device_name}.net_current",
        )
        self.current_label.displayFormat = DisplayFormat.Exponential
        self.current_label_layout.addWidget(self.current_label)
        self.current_unit_label = QLabel(parent=self.parent)
        self.current_unit_label.setText("A")
        self.current_label_layout.addWidget(self.current_unit_label)
        self.current_label_layout.addItem(HSpacer())
        self.column_layouts[2].addItem(VSpacer())
        # Label for the gain/offset column header
        self.column_layouts[3].addItem(VSpacer())
        self.gain_header_label = QLabel()
        self.gain_header_label.setText("Gain/Offset")
        self.gain_header_label.setAlignment(Qt.AlignCenter)
        self.column_layouts[3].addWidget(self.gain_header_label)
        # Gain up/down buttons
        self.gain_buttons_layout = QHBoxLayout()
        self.column_layouts[3].addLayout(self.gain_buttons_layout)
        self.gain_buttons_layout.addItem(HSpacer())
        self.gain_down_button = PyDMPushButton(
            init_channel=f"haven://{device_name}.preamp.gain_level",
            relative=True,
            pressValue=-1,
            icon=qta.icon("fa5s.arrow-left"),
        )
        self.gain_buttons_layout.addWidget(self.gain_down_button)
        self.gain_up_button = PyDMPushButton(
            init_channel=f"haven://{device_name}.preamp.gain_level",
            relative=True,
            pressValue=1,
            icon=qta.icon("fa5s.arrow-right"),
        )
        self.gain_buttons_layout.addWidget(self.gain_up_button)
        self.gain_buttons_layout.addItem(HSpacer())
        self.gain_monitor = PyDMChannel(
            address=f"haven://{device_name}.preamp.gain_level",
            value_slot=self.update_gain_level_widgets,
        )
        self.gain_monitor.connect()
        # Reporting the current gain as text
        self.gain_label_layout = QHBoxLayout()
        self.column_layouts[3].addLayout(self.gain_label_layout)
        self.gain_label_layout.addItem(HSpacer())
        self.gain_value_label = PyDMLabel(
            parent=self.parent,
            init_channel=f"haven://{device_name}.preamp.sensitivity_value",
        )
        self.gain_label_layout.addWidget(self.gain_value_label)
        self.gain_unit_label = PyDMLabel(
            parent=self.parent,
            init_channel=f"haven://{device_name}.preamp.sensitivity_unit",
        )
        self.gain_label_layout.addWidget(self.gain_unit_label)
        self.gain_label_layout.addItem(HSpacer())
        # Auto-gain and detail window controls
        self.column_layouts[4].addItem(VSpacer())
        self.auto_gain_checkbox = QCheckBox(parent=self.parent)
        self.auto_gain_checkbox.setText("Auto-gain")
        self.column_layouts[4].addWidget(self.auto_gain_checkbox)
        self.details_button = QPushButton(parent=self.parent)
        self.details_button.setText("More")
        self.details_button.setIcon(qta.icon("fa5s.cog"))
        self.details_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        self.column_layouts[4].addWidget(self.details_button)
        self.column_layouts[4].addItem(VSpacer())

    def update_gain_level_widgets(self, new_level):
        if new_level == 0:
            can_go_down, can_go_up = (False, True)
        elif new_level == 27:
            can_go_down, can_go_up = (True, False)
        else:
            can_go_down, can_go_up = (True, True)
        self.gain_down_button.setEnabled(can_go_down)
        self.gain_up_button.setEnabled(can_go_up)


class VSpacer(QSpacerItem):
    def __init__(
        self, w=40, h=20, hData=QSizePolicy.Expanding, vData=QSizePolicy.Minimum
    ):
        super().__init__(w, h, hData, vData)


class HSpacer(QSpacerItem):
    def __init__(
        self, w=40, h=20, hData=QSizePolicy.Expanding, vData=QSizePolicy.Minimum
    ):
        super().__init__(w, h, hData, vData)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging

import qtawesome as qta
from apsbss import apsbss
from dm.common.exceptions.objectNotFound import ObjectNotFound
from qtpy.QtCore import Signal
from qtpy.QtGui import QStandardItem, QStandardItemModel

from firefly import display
from haven import beamline, load_config

log = logging.getLogger(__name__)


class BssDisplay(display.FireflyDisplay):
    """A PyDM display for the beamline scheduling system (BSS)."""

    _proposal_col_names = ["ID", "Title", "Start", "End", "Users", "Badges"]
    _esaf_col_names = [
        "ID",
        "Title",
        "Start",
        "End",
        "Users",
        "Badges",
    ]
    _esaf_id: str = ""
    _psoporsal_id: str = ""

    # Signal
    proposal_changed = Signal()
    proposal_selected = Signal()
    esaf_changed = Signal()
    esaf_selected = Signal()

    def __init__(self, api=apsbss, args=None, macros={}, **kwargs):
        self.api = api
        super().__init__(args=args, macros=macros, **kwargs)
        # Load data models for proposals and ESAFs
        self.load_models()

    def customize_ui(self):
        super().customize_ui()
        icon = qta.icon("fa5s.arrow-right")
        self.ui.update_proposal_button.setIcon(icon)
        self.ui.update_proposal_button.clicked.connect(self.update_proposal)
        self.ui.update_esaf_button.setIcon(icon)
        self.ui.update_esaf_button.clicked.connect(self.update_esaf)
        self.ui.refresh_models_button.clicked.connect(self.load_models)
        # Icon for the refresh button
        self.ui.refresh_models_button.setIcon(qta.icon("fa5s.sync"))

    def customize_device(self):
        self._device = beamline.registry.find("beamline_manager")

    def proposals(self):
        config = load_config()
        proposals = []
        beamline = self._device.bss.proposal.beamline_name.get()
        cycle = self._device.bss.esaf.aps_cycle.get()
        # Get proposal data from the API
        try:
            api_result = self.api.listProposals(cycle, beamline)
        except ObjectNotFound:
            api_result = []
        # Parse the API payload into the format for the BSS IOC
        for proposal in api_result:
            users = proposal["experimenters"]
            proposals.append(
                {
                    "Title": proposal["title"],
                    "ID": proposal["id"],
                    "Start": proposal["startTime"],
                    "End": proposal["endTime"],
                    "Users": ", ".join([usr["lastName"] for usr in users]),
                    "Badges": ", ".join([usr["badge"] for usr in users]),
                }
            )
        return proposals

    def esafs(self):
        config = load_config()
        esafs_ = []
        beamline = self._device.bss.proposal.beamline_name.get()
        cycle = self._device.bss.esaf.aps_cycle.get()
        # Retrieve current ESAFS from data management API
        try:
            api_result = self.api.listESAFs(cycle, beamline.split("-")[0])
        except (ObjectNotFound, KeyError):
            api_result = []
        # Parse the API data into a format usable by the BSS IOC
        for esaf in api_result:
            users = esaf["experimentUsers"]
            esafs_.append(
                {
                    "Title": esaf["esafTitle"],
                    "ID": esaf["esafId"],
                    "Start": esaf["experimentStartDate"],
                    "End": esaf["experimentEndDate"],
                    "Users": ", ".join([usr["lastName"] for usr in users]),
                    "Badges": ", ".join([usr["badge"] for usr in users]),
                }
            )
        return esafs_

    def load_models(self):
        config = load_config()
        # Create proposal model object
        col_names = self._proposal_col_names
        self.proposal_model = QStandardItemModel()
        self.proposal_model.setHorizontalHeaderLabels(col_names)
        # Load individual proposals
        proposals = self.proposals()
        for proposal in proposals:
            items = [QStandardItem(str(proposal[col])) for col in col_names]
            self.proposal_model.appendRow(items)
        self.ui.proposal_view.setModel(self.proposal_model)
        # Create proposal model object
        col_names = self._esaf_col_names
        self.esaf_model = QStandardItemModel()
        self.esaf_model.setHorizontalHeaderLabels(col_names)
        # Load individual esafs
        esafs = self.esafs()
        for esaf in esafs:
            items = [QStandardItem(str(esaf[col])) for col in col_names]
            self.esaf_model.appendRow(items)
        self.ui.esaf_view.setModel(self.esaf_model)
        # Connect slots for when proposal/ESAF is changed
        self.ui.proposal_view.selectionModel().currentChanged.connect(
            self.select_proposal
        )
        self.ui.esaf_view.selectionModel().currentChanged.connect(self.select_esaf)

    def select_proposal(self, current, previous):
        # Determine which proposal was selected
        id_col_idx = self._proposal_col_names.index("ID")
        new_id = current.siblingAtColumn(id_col_idx).data()
        self._proposal_id = new_id
        # Enable controls for updating the metadata
        self.ui.update_proposal_button.setEnabled(True)
        self.proposal_selected.emit()

    def update_proposal(self):
        new_id = self._proposal_id
        # Change the proposal in the EPICS record
        bss = beamline.registry.find("beamline_manager.bss")
        bss.proposal.proposal_id.set(new_id).wait()
        # Notify any interested parties that the proposal has been changed
        self.proposal_changed.emit()

    def select_esaf(self, current, previous):
        # Determine which esaf was selected
        id_col_idx = self._esaf_col_names.index("ID")
        new_id = current.siblingAtColumn(id_col_idx).data()
        self._esaf_id = new_id
        # Enable controls for updating the metadata
        self.ui.update_esaf_button.setEnabled(True)
        self.esaf_selected.emit()

    def update_esaf(self):
        new_id = self._esaf_id
        # Change the esaf in the EPICS record
        bss = beamline.registry.find("beamline_manager.bss")
        bss.wait_for_connection()
        bss.esaf.esaf_id.set(new_id).wait(timeout=5)
        # Notify any interested parties that the esaf has been changed
        self.esaf_changed.emit()

    def ui_filename(self):
        return "bss.ui"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""
test the SRS DG-645 digital delay device support

Hardware is not available so test with best efforts
"""

from haven.devices import delay


async def test_dg645_device():
    dg645 = delay.DG645Delay("", name="delay")
    await dg645.connect(mock=True)
    read_names = []
    read_attrs = (await dg645.describe()).keys()
    assert sorted(read_attrs) == read_names

    cfg_names = [
        "delay-burst_config",
        "delay-burst_count",
        "delay-burst_delay",
        "delay-burst_mode",
        "delay-burst_period",
        "delay-channels-A-reference",
        "delay-channels-A-delay",
        "delay-channels-B-reference",
        "delay-channels-B-delay",
        "delay-channels-C-reference",
        "delay-channels-C-delay",
        "delay-channels-D-reference",
        "delay-channels-D-delay",
        "delay-channels-E-reference",
        "delay-channels-E-delay",
        "delay-channels-F-reference",
        "delay-channels-F-delay",
        "delay-channels-G-reference",
        "delay-channels-G-delay",
        "delay-channels-H-reference",
        "delay-channels-H-delay",
        "delay-device_id",
        "delay-label",
        "delay-outputs-AB-amplitude",
        "delay-outputs-AB-offset",
        "delay-outputs-AB-polarity",
        "delay-outputs-AB-trigger_phase",
        "delay-outputs-AB-trigger_prescale",
        "delay-outputs-CD-amplitude",
        "delay-outputs-CD-offset",
        "delay-outputs-CD-polarity",
        "delay-outputs-CD-trigger_phase",
        "delay-outputs-CD-trigger_prescale",
        "delay-outputs-EF-amplitude",
        "delay-outputs-EF-offset",
        "delay-outputs-EF-polarity",
        "delay-outputs-EF-trigger_phase",
        "delay-outputs-EF-trigger_prescale",
        "delay-outputs-GH-amplitude",
        "delay-outputs-GH-offset",
        "delay-outputs-GH-polarity",
        "delay-outputs-GH-trigger_phase",
        "delay-outputs-GH-trigger_prescale",
        "delay-outputs-T0-amplitude",
        "delay-outputs-T0-offset",
        "delay-outputs-T0-polarity",
        "delay-trigger_advanced_mode",
        "delay-trigger_holdoff",
        "delay-trigger_inhibit",
        "delay-trigger_level",
        "delay-trigger_prescale",
        "delay-trigger_rate",
        "delay-trigger_source",
    ]
    cfg_attrs = (await dg645.describe_configuration()).keys()
    assert sorted(cfg_attrs) == sorted(cfg_names)

    # List all the components
    cpt_names = [
        "delay-autoip_state",
        "delay-bare_socket_state",
        "delay-burst_config",
        "delay-burst_count",
        "delay-burst_delay",
        "delay-burst_mode",
        "delay-burst_period",
        "delay-channels",
        "delay-clear_error",
        "delay-device_id",
        "delay-dhcp_state",
        "delay-gateway",
        "delay-goto_local",
        "delay-goto_remote",
        "delay-gpib_address",
        "delay-gpib_state",
        "delay-ip_address",
        "delay-label",
        "delay-lan_state",
        "delay-mac_address",
        "delay-network_mask",
        "delay-outputs",
        "delay-reset",
        "delay-reset_gpib",
        "delay-reset_lan",
        "delay-reset_serial",
        "delay-serial_baud",
        "delay-serial_state",
        "delay-static_ip_state",
        "delay-status",
        "delay-status_checking",
        "delay-telnet_state",
        "delay-trigger_advanced_mode",
        "delay-trigger_holdoff",
        "delay-trigger_inhibit",
        "delay-trigger_level",
        "delay-trigger_prescale",
        "delay-trigger_rate",
        "delay-trigger_source",
        "delay-vxi11_state",
    ]
    child_names = [child.name for attr, child in dg645.children()]
    assert sorted(child_names) == sorted(cpt_names)



from pathlib import Path
from unittest.mock import MagicMock

from firefly.action import WindowAction


def test_create_window(qtbot):
    action = WindowAction(
        name="start_queue",
        text="Start queue",
        display_file=Path(),
        WindowClass=MagicMock,
    )
    with qtbot.waitSignal(action.window_created):
        action.create_window()



import logging
from enum import Enum

from ophyd_async.core import ConfigSignal, StandardReadable
from ophyd_async.epics.signal import epics_signal_rw

from .motor import Motor

log = logging.getLogger(__name__)


class Monochromator(StandardReadable):
    _ophyd_labels_ = {"monochromators"}

    class Mode(str, Enum):
        FIXED_OFFSET = "Si(111) Fixed Offset"
        CHANNEL_CUT = "Si(111) Channel-cut"
        ML48 = "Multi-layer 4.8nm"
        ML24 = "Multi-layer 2.4nm"

    def __init__(self, prefix: str, name: str = ""):
        with self.add_children_as_readables():
            # Virtual motors
            self.energy = Motor(f"{prefix}Energy")
            self.offset = Motor(f"{prefix}Offset")
            # ACS Motors
            self.bragg = Motor(f"{prefix}ACS:m3")
            self.gap = Motor(f"{prefix}ACS:m4")
            self.horiz = Motor(f"{prefix}ACS:m1")
            self.vert = Motor(f"{prefix}ACS:m2")
            self.roll2 = Motor(f"{prefix}ACS:m5")
            self.pitch2 = Motor(f"{prefix}ACS:m6")
        with self.add_children_as_readables(ConfigSignal):
            # Transform constants, etc.
            self.id_tracking = epics_signal_rw(bool, f"{prefix}ID_tracking")
            self.id_offset = epics_signal_rw(float, f"{prefix}ID_offset")
            self.d_spacing = epics_signal_rw(float, f"{prefix}dspacing")
            self.d_spacing_unit = epics_signal_rw(str, f"{prefix}dspacing.EGU")
            self.mode = epics_signal_rw(self.Mode, f"{prefix}mode")
            self.energy_constant1 = epics_signal_rw(float, f"{prefix}EnergyC1.VAL")
            self.energy_constant2 = epics_signal_rw(float, f"{prefix}EnergyC2.VAL")
            self.energy_constant3 = epics_signal_rw(float, f"{prefix}EnergyC3.VAL")
            # Interferomters
            # self.roll_int = Motor(f"{prefix}ACS:m7")
            # self.pi_int = Motor(f"{prefix}ACS:m8")
        super().__init__(name=name)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import subprocess
from pathlib import Path
from typing import Optional, Sequence

from ophyd import Device
from pydm import Display
from qtpy import QtWidgets
from qtpy.QtCore import Signal, Slot

from haven import beamline


class FireflyDisplay(Display):
    caqtdm_ui_file: str = ""
    caqtdm_command: str = "/APSshare/bin/caQtDM -style plastique -noMsg -attach"
    caqtdm_actions: Sequence
    device: Optional[Device]
    registry = None

    # Signals
    status_message_changed = Signal(str, int)
    queue_item_submitted = Signal(object)

    def __init__(self, parent=None, args=None, macros=None, ui_filename=None, **kwargs):
        super().__init__(
            parent=parent, args=args, macros=macros, ui_filename=ui_filename, **kwargs
        )
        self.customize_device()
        self.customize_ui()
        self.prepare_caqtdm_actions()

    def prepare_caqtdm_actions(self):
        """Create QActions for opening caQtDM panels.

        By default, this method creates one action if
        *self.caqtdm_ui_file* is set. Individual displays should
        override this method to add their own QActions. Any actions
        added to the *self.caqtdm_actions* list will be added to the
        "Setup" menu if the display is the root display in a main
        window.

        """
        self.caqtdm_actions = []
        if self.caqtdm_ui_file != "":
            # Create an action for launching a single caQtDM file
            action = QtWidgets.QAction(self)
            action.setObjectName("launch_caqtdm_action")
            action.setText("ca&QtDM")
            action.triggered.connect(self.launch_caqtdm)
            try:
                tooltip = f"Launch the caQtDM panel for {self.device.name}"
            except AttributeError:
                tooltip = "Launch the caQtDM panel for this display."
            action.setToolTip(tooltip)
            self.caqtdm_actions.append(action)

    def _all_children(self, widget):
        for child in widget.children():
            yield widget
            yield from self._all_children(widget=child)

    def find_plan_widgets(self):
        """Look through widgets and determine if any of them are used for
        bluesky plans.

        """
        # from pprint import pprint
        # pprint([c.objectName() for c in self._all_children(self)])
        # for child in self.ui.children():
        #     if child.objectName() == "set_energy_button":
        #         print(f"**{child.objectName()}**")
        #     else:
        #         print(child.objectName())

    def _open_caqtdm_subprocess(self, cmds, *args, **kwargs):
        """Launch a new subprocess and save it to self._caqtdm_process."""
        # Try to leave this as just a simple call to Popen.
        # It helps simplify testing
        self._caqtdm_process = subprocess.Popen(cmds, *args, **kwargs)

    @Slot()
    def launch_caqtdm(self, macros={}, ui_file: str = None):
        """Launch a caQtDM window showing the window's panel."""
        if ui_file is None:
            ui_file = self.caqtdm_ui_file
        cmds = self.caqtdm_command.split()
        # Add macros
        macro_str = ",".join(f"{key}={val}" for key, val in macros.items())
        if macro_str != "":
            cmds.extend(["-macro", macro_str])
        # Add the path to caQtDM .ui file
        cmds.append(ui_file)
        self._open_caqtdm_subprocess(cmds)

    async def update_devices(self, registry):
        """The list of accessible devices has changed."""
        self.registry = registry

    def customize_device(self):
        # Retrieve the device
        device = self.macros().get("DEVICE")
        if device is not None:
            device = beamline.registry.find(device)
        self.device = device
        return device

    def customize_ui(self):
        pass

    def update_queue_status(self, status):
        pass

    def show_message(self, message, timeout=0):
        """Display a message in the status bar."""
        self.status_message_changed.emit(str(message), timeout)

    def ui_filename(self):
        raise NotImplementedError

    def ui_filepath(self):
        path_base = Path(__file__).parent
        return path_base / self.ui_filename()


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from haven.devices.shutter import ShutterState
from haven.plans.record_dark_current import record_dark_current


def test_shutters_get_reset(shutters, ion_chamber):
    shutter = shutters[0]
    msgs = list(record_dark_current(ion_chambers=[ion_chamber], shutters=[shutter]))
    # Check the shutters get closed
    set_shutter_msg = msgs[1]
    assert set_shutter_msg.command == "set"
    assert set_shutter_msg.obj is shutter
    assert set_shutter_msg.args[0] == ShutterState.CLOSED
    # Check the shutters get re-opened
    set_shutter_msg = msgs[-2]
    assert set_shutter_msg.command == "set"
    assert set_shutter_msg.obj is shutter
    assert set_shutter_msg.args[0] == ShutterState.OPEN


def test_messages(shutters, ion_chamber):
    shutter = shutters[0]
    msgs = list(record_dark_current(ion_chambers=[ion_chamber], shutters=[shutter]))
    from pprint import pprint

    pprint(msgs)
    # Check the shutters get closed
    trigger_msg = msgs[3]
    assert trigger_msg.obj is ion_chamber
    assert trigger_msg.kwargs["record_dark_current"] is True



from queue import Queue
from unittest.mock import MagicMock

import numpy as np
import pytest
from bluesky_adaptive.recommendations import NoRecommendation

from haven import GainRecommender, auto_gain
from haven.plans import auto_gain as auto_gain_module


def test_plan_recommendations(ion_chamber):
    # Make a fake queue that accepts the second step
    queue = Queue()
    queue.put({ion_chamber.preamp.gain_level.name: 12})
    queue.put(None)
    # Prepare the plan and make sure it generates messages
    plan = auto_gain(ion_chambers=[ion_chamber], queue=queue)
    msgs = list(plan)
    # Make sure the plan sets the gain values properly
    set_msgs = [msg for msg in msgs if msg.command == "set"]
    assert len(set_msgs) == 2
    assert set_msgs[0].args[0] == 13  # Starting point
    assert set_msgs[1].args[0] == 12  # First adaptive point
    # Make sure the plan triggers the ion chamber
    trigger_msgs = [msg for msg in msgs if msg.command == "trigger"]
    assert len(trigger_msgs) == 2


@pytest.mark.parametrize(
    "prefer,target_volts",
    [("middle", 2.5), ("lower", 0.5), ("upper", 4.5)],
)
async def test_plan_prefer_arg(ion_chamber, monkeypatch, prefer, target_volts):
    """Check that the *prefer* argument works properly."""
    queue = Queue()
    queue.put({ion_chamber.preamp.gain_level.name: 12})
    queue.put(None)
    monkeypatch.setattr(auto_gain_module, "GainRecommender", MagicMock())
    plan = auto_gain(ion_chambers=[ion_chamber], queue=queue, prefer=prefer)
    msgs = list(plan)
    auto_gain_module.GainRecommender.assert_called_with(
        volts_min=0.5, volts_max=4.5, target_volts=target_volts
    )


@pytest.fixture()
def recommender():
    recc = GainRecommender()
    return recc


def test_recommender_all_low(recommender):
    """Does the engine get the right direction when all pre-amp gains are
    too low?

    """
    # Gain sensitivity levels
    gains = np.asarray([[10, 13]])
    assert gains.shape == (1, 2)
    # Corresponding volts
    volts = np.asarray([[0.1, 0.25]])
    assert volts.shape == (1, 2)
    # Check recommendations are lower by 3
    recommender.tell_many(gains, volts)
    np.testing.assert_equal(recommender.ask(1), (13, 16))


def test_recommender_some_low(recommender):
    """Does the engine get the direction right when some pre-amp gains are
    too low and some are just right?

    """
    # Gain sensitivity levels
    gains = np.asarray([[10, 13]])
    assert gains.shape == (1, 2)
    # Corresponding volts
    volts = np.asarray([[0.1, 2.5]])
    assert volts.shape == (1, 2)
    # Check recommendations
    recommender.tell_many(gains, volts)
    np.testing.assert_equal(recommender.ask(1), (13, 14))


def test_recommender_all_high(recommender):
    """Does the engine get the direction correct when all pre-amp gains
    are too high?

    """
    # Gain sensitivity levels
    gains = np.asarray([[10, 13]])
    assert gains.shape == (1, 2)
    recommender.last_point = np.asarray([10, 13])
    # Corresponding volts
    volts = np.asarray([[4.6, 5.2]])
    assert volts.shape == (1, 2)
    # Check recommendations
    recommender.tell_many(gains, volts)
    np.testing.assert_equal(recommender.ask(1), (7, 10))


def test_recommender_some_high(recommender):
    # Gain sensitivity levels
    gains = np.asarray([[10, 13]])
    assert gains.shape == (1, 2)
    # Corresponding volts
    volts = np.asarray([[5.7, 2.5]])
    assert volts.shape == (1, 2)
    # Check recommendations
    recommender.tell_many(gains, volts)
    np.testing.assert_equal(recommender.ask(1), (7, 14))


def test_recommender_high_and_low(recommender):
    # Gain sensitivity levels
    gains = np.asarray([[10, 13]])
    assert gains.shape == (1, 2)
    # Corresponding volts
    volts = np.asarray([[5.7, 0.23]])
    assert volts.shape == (1, 2)
    # Check recommendations
    recommender.tell_many(gains, volts)
    np.testing.assert_equal(recommender.ask(1), (7, 16))


def test_recommender_fill_missing_gains(recommender):
    """If we have both high and low gains already done, will the
    recommender fill in the missing pieces."""
    gains = [
        [9],
        [12],
        [10],
    ]
    volts = [
        [0.3],
        [5.2],
        [1.5],
    ]
    recommender.tell_many(gains, volts)
    # Does it recommend the missing gain value?
    df = recommender.dfs[0]
    assert recommender.next_gain(df) == 11


def test_recommender_no_high(recommender):
    """If the gain happens to be in range first try, we need to get
    more data to be sure we have the best point.

    """
    gains = [[8], [9], [10]]
    volts = [[0.3], [1.25], [2.7]]
    recommender.tell_many(gains, volts)
    df = recommender.dfs[0]
    assert recommender.next_gain(df) == 11


def test_recommender_no_low(recommender):
    """If the gain happens to be in range first try, we need to get
    more data to be sure we have the best point.

    """
    gains = [[8], [9], [10]]
    volts = [[1.25], [2.75], [5.4]]
    recommender.tell_many(gains, volts)
    df = recommender.dfs[0]
    assert recommender.next_gain(df) == 7


def test_recommender_no_solution(recommender):
    """If the gain profile goes from too low to too high in one step, what should we report?"""
    gains = [[9], [10]]
    volts = [[0.3], [5.2]]
    recommender.tell_many(gains, volts)
    # Does it recommend the missing gain value?
    df = recommender.dfs[0]
    assert recommender.next_gain(df) == 9


@pytest.mark.parametrize("target_volts,gain", [(0.5, 10), (2.5, 9), (4.5, 8)])
def test_recommender_correct_solution(target_volts, gain):
    """If the gain profile goes from too low to too high in one step, what should we report?"""
    recommender = GainRecommender(target_volts=target_volts)
    gains = [[7], [8], [9], [10], [11]]
    volts = [[5.2], [4.1], [2.7], [1.25], [0.4]]
    recommender.tell_many(gains, volts)
    # Does it recommend the missing gain value?
    df = recommender.dfs[0]
    assert recommender.next_gain(df) == gain


def test_recommender_gain_range_high(recommender):
    """Check that the recommender doesn't go outside the allowed range."""
    gains = [[27]]
    volts = [[0.3]]
    recommender.tell_many(gains, volts)
    # Does it recommend the missing gain value?
    df = recommender.dfs[0]
    assert recommender.next_gain(df) == 27


def test_recommender_gain_range_low(recommender):
    """Check that the recommender doesn't go outside the allowed range."""
    gains = [[0]]
    volts = [[5.4]]
    recommender.tell_many(gains, volts)
    # Does it recommend the missing gain value?
    df = recommender.dfs[0]
    assert recommender.next_gain(df) == 0


def test_recommender_no_change(recommender):
    """If we've already found the best solution, we should stop the scan."""
    # Set a history of gain measurements
    gains = np.asarray(
        [
            [10, 13],
            [11, 11],
            [12, 14],
            [13, 12],
        ]
    )
    volts = np.asarray(
        [
            [7.10, 2.23],
            [4.30, 5.30],
            [2.10, 0.10],
            [0.20, 4.33],
        ]
    )
    recommender.tell_many(gains, volts)
    # Make sure the engine recommends the best solution
    assert recommender.ask(1) == [12, 13]
    # Pass in results for this new measurement
    recommender.tell_many([[12, 13]], [[2.13, 2.19]])
    # Check recommendations
    with pytest.raises(NoRecommendation):
        recommender.ask(1)



import pytest
from ophyd import Device, EpicsMotor, sim
from ophydregistry import Registry

from haven.exceptions import ComponentNotFound, MultipleComponentsFound


@pytest.fixture()
def reg():
    yield Registry(auto_register=False)


def test_register_component(reg):
    # Create an unregistered component
    cpt = sim.SynGauss(
        "I0",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    # Make sure the component doesn't get found without being registered
    with pytest.raises(ComponentNotFound):
        list(reg.findall(label="ion_chamber"))
    with pytest.raises(ComponentNotFound):
        list(reg.findall(name="I0"))
    # Now register the component
    cpt = reg.register(cpt)
    # Confirm that it's findable by label
    results = reg.findall(label="ion_chamber")
    assert cpt in results
    # Config that it's findable by name
    results = reg.findall(name="I0")
    assert cpt in results


def test_find_missing_components(reg):
    """Test that registry raises an exception if no matches are found."""
    cpt = sim.SynGauss(
        "I0",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    reg.register(cpt)
    # Now make sure a different query still returns no results
    with pytest.raises(ComponentNotFound):
        reg.findall(label="spam")


def test_find_allow_missing_components(reg):
    """Test that registry tolerates missing components with the
    *allow_none* argument.

    """
    # Get some non-existent devices and check that the right nothing is returned
    assert list(reg.findall(label="spam", allow_none=True)) == []
    assert reg.find(name="eggs", allow_none=True) is None


def test_exceptions(reg):
    reg.register(Device("", name="It"))
    # Test if a non-existent labels throws an exception
    with pytest.raises(ComponentNotFound):
        reg.find(label="spam")


def test_as_class_decorator(reg):
    # Create a dummy decorated class
    IonChamber = type("IonChamber", (Device,), {})
    IonChamber = reg.register(IonChamber)
    # Instantiate the class
    IonChamber("PV_PREFIX", name="I0", labels={"ion_chamber"})
    # Check that it gets retrieved
    result = reg.find(label="ion_chamber")
    assert result.prefix == "PV_PREFIX"
    assert result.name == "I0"


def test_find_component(reg):
    # Create an unregistered component
    cptA = sim.SynGauss(
        "I0",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    cptB = sim.SynGauss(
        "It",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    # Register the components
    reg.register(cptA)
    reg.register(cptB)
    # Only one match should work fine
    result = reg.find(name="I0")
    assert result is cptA
    # Multiple matches should raise an exception
    with pytest.raises(MultipleComponentsFound):
        result = reg.find(label="ion_chamber")


def test_find_name_by_dot_notation(reg):
    # Create a simulated component
    cptA = sim.SynGauss(
        "I0",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    reg.register(cptA)
    # Only one match should work fine
    result = reg.find(name="I0.val")
    assert result is cptA.val


def test_find_labels_by_dot_notation(reg):
    # Create a simulated component
    cptA = sim.SynGauss(
        "I0",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    reg.register(cptA)
    # Only one match should work fine
    result = reg.find(label="ion_chamber.val")
    assert result is cptA.val


def test_find_any(reg):
    # Create an unregistered component
    cptA = sim.SynGauss(
        "I0",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    cptB = sim.SynGauss(
        "It",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={"ion_chamber"},
    )
    cptC = sim.SynGauss(
        "ion_chamber",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
    )
    cptD = sim.SynGauss(
        "sample motor",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={},
    )
    # Register the components
    reg.register(cptA)
    reg.register(cptB)
    reg.register(cptC)
    reg.register(cptD)
    # Only one match should work fine
    result = reg.findall(any_of="ion_chamber")
    assert cptA in result
    assert cptB in result
    assert cptC in result
    assert cptD not in result


def test_find_by_device(reg):
    """The registry should just return the device itself if that's what is passed."""
    # Register a component
    cptD = sim.SynGauss(
        "sample motor",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={},
    )
    reg.register(cptD)
    # Pass the device itself to the find method
    result = reg.find(cptD)
    assert result is cptD


def test_find_by_list_of_names(reg):
    """Will the findall() method handle lists of things to look up."""
    # Register a component
    cptA = sim.SynGauss(
        "sample motor A",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={},
    )
    cptB = sim.SynGauss(
        "sample motor B",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={},
    )
    cptC = sim.SynGauss(
        "sample motor C",
        sim.motor,
        "motor",
        center=-0.5,
        Imax=1,
        sigma=1,
        labels={},
    )
    reg.register(cptA)
    reg.register(cptB)
    reg.register(cptC)
    # Pass the device names into the findall method
    result = reg.findall(["sample motor A", "sample motor B"])
    assert cptA in result
    assert cptB in result
    assert cptC not in result


def test_user_readback(reg):
    """Edge case where EpicsMotor.user_readback is named the same as the motor itself."""
    device = EpicsMotor("", name="epics_motor")
    reg.register(device)
    # See if requesting the device.user_readback returns the proper signal
    reg["epics_motor_user_readback"]


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""Ophyd device support for a set of XIA PFCU-controlled filters.

A PFCUFilterBank controls a set of 4 filters. Optionally, 2 filters in
a filter bank can be used as a shutter.

"""

from enum import IntEnum
from typing import Sequence

from ophyd import Component as Cpt
from ophyd import DynamicDeviceComponent as DCpt
from ophyd import EpicsSignal, EpicsSignalRO
from ophyd import FormattedComponent as FCpt
from ophyd import PVPositionerIsClose
from ophyd.signal import DerivedSignal

from .shutter import ShutterState


class FilterPosition(IntEnum):
    OUT = 0
    IN = 1


class PFCUFilter(PVPositionerIsClose):
    """A single filter in a PFCU filter bank.

    E.g. 25idc:pfcu0:filter1_mat
    """

    material = Cpt(EpicsSignal, "_mat", kind="config")
    thickness = Cpt(EpicsSignal, "_thick", kind="config")
    thickness_unit = Cpt(EpicsSignal, "_thick.EGU", kind="config")
    notes = Cpt(EpicsSignal, "_other", kind="config")

    setpoint = Cpt(EpicsSignal, "", kind="normal")
    readback = Cpt(EpicsSignalRO, "_RBV", kind="normal")


shutter_state_map = {
    # (top filter, bottom filter): state
    (FilterPosition.OUT, FilterPosition.IN): ShutterState.OPEN,
    (FilterPosition.IN, FilterPosition.OUT): ShutterState.CLOSED,
    (FilterPosition.OUT, FilterPosition.OUT): ShutterState.FAULT,
    (FilterPosition.IN, FilterPosition.IN): ShutterState.FAULT,
}


class PFCUShutterSignal(DerivedSignal):
    def _mask(self, pos):
        num_filters = 4
        return 1 << (num_filters - pos)

    def top_mask(self):
        return self._mask(self.parent._top_filter)

    def bottom_mask(self):
        return self._mask(self.parent._bottom_filter)

    def forward(self, value):
        """Convert shutter state to filter bank state."""
        # Bit masking to set both blades together
        old_bits = self.derived_from.parent.readback.get(as_string=False)
        if value == ShutterState.OPEN:
            open_bits = self.bottom_mask()
            close_bits = self.top_mask()
        elif value == ShutterState.CLOSED:
            close_bits = self.bottom_mask()
            open_bits = self.top_mask()
        else:
            raise ValueError(bin(value))
        new_bits = (old_bits | open_bits) & (0b1111 - close_bits)
        return new_bits

    def inverse(self, value):
        """Convert filter bank state to shutter state."""
        # Determine which filters are open and closed
        top_position = int(bool(value & self.top_mask()))
        bottom_position = int(bool(value & self.bottom_mask()))
        result = shutter_state_map[(top_position, bottom_position)]
        return result


class PFCUShutter(PVPositionerIsClose):
    """A shutter made of two PFCU4 filters.

    For faster operation, both filters will be moved at the same
    time. That means this shutter must be a sub-component of a
    :py:class:`PFCUFilterBank`.

    Parameters
    ==========
    top_filter
      The PV for the filter that is open when the filter is set to
      "out".
    bottom_filter
      The PV for the filter that is open when the filter is set to
      "in".

    """

    readback = Cpt(PFCUShutterSignal, derived_from="parent.parent.readback")
    setpoint = Cpt(PFCUShutterSignal, derived_from="parent.parent.setpoint")

    top_filter = FCpt(PFCUFilter, "{self.prefix}filter{self._top_filter}")
    bottom_filter = FCpt(PFCUFilter, "{self.prefix}filter{self._bottom_filter}")

    def __init__(
        self,
        prefix: str = "",
        *,
        name: str,
        top_filter: str,
        bottom_filter: str,
        labels={"shutters"},
        **kwargs,
    ):
        self._top_filter = top_filter
        self._bottom_filter = bottom_filter
        super().__init__(
            prefix=prefix,
            name=name,
            limits=(ShutterState.OPEN, ShutterState.CLOSED),
            labels=labels,
            **kwargs,
        )


class PFCUFilterBank(PVPositionerIsClose):
    """Parameters
    ==========
    shutters
      Sets of filter numbers to use as shutters. Each entry in
      *shutters* should be a tuple like (top, bottom) where the first
      filter (top) is open when the filter is set to "out".

    """

    num_slots: int = 4

    readback = Cpt(EpicsSignalRO, "config_RBV", kind="normal")
    setpoint = Cpt(EpicsSignal, "config", kind="normal")

    def __new__(cls, prefix: str, name: str, shutters=[], **kwargs):
        # Determine which filters to use as filters vs shutters
        all_shutters = [v for shutter in shutters for v in shutter]
        filters = [
            idx for idx in range(1, cls.num_slots + 1) if idx not in all_shutters
        ]
        # Create device components for filters and shutters
        comps = {
            "shutters": DCpt(
                {
                    f"shutter_{idx}": (
                        PFCUShutter,
                        "",
                        {
                            "top_filter": top,
                            "bottom_filter": bottom,
                            "labels": {"shutters"},
                        },
                    )
                    for idx, (top, bottom) in enumerate(shutters)
                }
            ),
            "filters": DCpt(
                {
                    f"filter{idx}": (
                        PFCUFilter,
                        f"filter{idx}",
                        {"labels": {"filters"}},
                    )
                    for idx in filters
                }
            ),
        }
        # Create any new child class with shutters and filters
        new_cls = type(cls.__name__, (PFCUFilterBank,), comps)
        return super().__new__(new_cls)

    def __init__(
        self,
        prefix: str = "",
        *,
        name: str,
        shutters: Sequence = [],
        labels: str = {"filter_banks"},
        **kwargs,
    ):
        super().__init__(prefix=prefix, name=name, labels=labels, **kwargs)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging

import numpy as np
import pytest

from haven import ERange, KRange, merge_ranges

# import unittest


logging.basicConfig(level=logging.INFO)

energy_range_parameters = [
    # (start, end,  step, expected_energies)
    (8300, 8400, 0.5, np.linspace(8300, 8400, num=201)),  # Example values
    (1, 1.3, 0.1, [1, 1.1, 1.2, 1.3]),  # Known float rounding error
    (0, 70, 50, [0, 50]),  # End-point is not a step multiple
    (1.3, 1.0, -0.1, [1.3, 1.2, 1.1, 1.0]),  # Reverse direction
]


@pytest.mark.parametrize("start,stop,step,expected_energies", energy_range_parameters)
def test_e_range(start, stop, step, expected_energies):
    """Test the ERange class for calculating energy points."""
    e_range = ERange(E_min=start, E_max=stop, E_step=step, exposure=0.1)
    np.testing.assert_allclose(e_range.energies(), expected_energies)
    np.testing.assert_allclose(e_range.exposures(), [0.1] * len(expected_energies))


def test_k_range():
    E0 = 17038
    k_range = KRange(k_min=2.8, k_max=14, k_step=0.05, k_weight=1, exposure=1.0)

    np.testing.assert_almost_equal(k_range.wavenumbers()[0], 2.8)
    np.testing.assert_almost_equal(k_range.wavenumbers()[-1], 14)
    energies = k_range.energies() + E0

    # Verify the results compared to Shelly's spreadsheet
    E_min = k_range.energies()[0]
    np.testing.assert_equal(energies[0], E0 + E_min)
    np.testing.assert_almost_equal(
        energies[-1], 746.7568692 + 17038, decimal=3
    )  # Athena value
    np.testing.assert_equal(k_range.exposures()[0], 1.0)
    np.testing.assert_almost_equal(k_range.exposures()[16], 1.28529317348436, decimal=3)


def test_merge_ranges():
    e_range1 = ERange(1, 5, 1, exposure=0.5)
    e_range2 = ERange(5, 7, 0.5, exposure=1)
    merged, exposures = merge_ranges(e_range2, e_range1)
    # Test validity of the result
    np.testing.assert_equal(merged, [1, 2, 3, 4, 5, 5.5, 6, 6.5, 7])
    np.testing.assert_equal(exposures, [0.5, 0.5, 0.5, 0.5, 1.0, 1.0, 1.0, 1.0, 1.0])


def test_merge_ranges_sorting():
    e_range1 = ERange(5, 7, 0.5, exposure=1)
    e_range2 = ERange(1, 5, 1, exposure=0.5)
    merged, exposures = merge_ranges(e_range2, e_range1, sort=True)
    # Test validity of the result
    np.testing.assert_equal(merged, [1, 2, 3, 4, 5, 5.5, 6, 6.5, 7])
    np.testing.assert_equal(exposures, [0.5, 0.5, 0.5, 0.5, 0.5, 1.0, 1.0, 1.0, 1.0])


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import warnings
from enum import IntEnum, unique

from ophyd import Component as Cpt
from ophyd import EpicsSignal, EpicsSignalRO
from ophyd.pv_positioner import PVPositionerIsClose
from ophyd.utils.errors import ReadOnlyError
from pcdsdevices.signal import MultiDerivedSignal
from pcdsdevices.type_hints import SignalToValue

# from apstools.devices.shutters import ApsPssShutterWithStatus as Shutter


log = logging.getLogger(__name__)


@unique
class ShutterState(IntEnum):
    OPEN = 0  # 0b000
    CLOSED = 1  # 0b001
    FAULT = 3  # 0b011
    UNKNOWN = 4  # 0b100


class PssShutter(PVPositionerIsClose):
    _last_setpoint: int = ShutterState.UNKNOWN
    allow_open: bool
    allow_close: bool

    def __init__(
        self,
        prefix: str,
        name: str,
        allow_open: bool = True,
        allow_close: bool = True,
        labels={"shutters"},
        **kwargs,
    ):
        self.allow_open = allow_open
        self.allow_close = allow_close
        super().__init__(prefix=prefix, name=name, labels=labels, **kwargs)

    def check_value(self, pos):
        """Check that the shutter has the right permissions."""
        if pos == ShutterState.CLOSED and not self.allow_close:
            raise ReadOnlyError(
                f"Shutter {self.name} is not permitted to be closed. Set `allow_close` for this shutter."
            )
        if pos == ShutterState.OPEN and not self.allow_open:
            raise ReadOnlyError(
                f"Shutter {self.name} is not permitted to be opened per iconfig.toml. Set `allow_open` for this shutter."
            )

    def _actuate_shutter(self, mds: MultiDerivedSignal, value: int) -> SignalToValue:
        """Open/close the shutter using derived-from signals."""
        if value == ShutterState.OPEN:
            items = {self.open_signal: 1}
        elif value == ShutterState.CLOSED:
            items = {self.close_signal: 1}
        else:
            raise ValueError(f"Invalid shutter state for {self}")
        return items

    def _shutter_setpoint(self, mds: MultiDerivedSignal, items: SignalToValue) -> int:
        """Determine whether the shutter was last opened or closed."""
        do_open = items[self.open_signal]
        do_close = items[self.close_signal]
        if do_open and do_close:
            # Shutter is both opening and closing??
            warnings.warn("Unknown shutter setpoint")
            self._last_setpoint = ShutterState.UNKNOWN
        elif do_open:
            self._last_setpoint = ShutterState.OPEN
        elif do_close:
            self._last_setpoint = ShutterState.CLOSED
        return self._last_setpoint

    readback = Cpt(EpicsSignalRO, "BeamBlockingM.VAL")
    setpoint = Cpt(
        MultiDerivedSignal,
        attrs=["open_signal", "close_signal"],
        calculate_on_put=_actuate_shutter,
        calculate_on_get=_shutter_setpoint,
    )
    open_signal = Cpt(EpicsSignal, "OpenEPICSC", kind="omitted")
    close_signal = Cpt(EpicsSignal, "CloseEPICSC", kind="omitted")


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import json
from typing import Sequence

from pydm.widgets import PyDMEmbeddedDisplay

from firefly import display
from haven import beamline


class FiltersDisplay(display.FireflyDisplay):
    filters: Sequence

    def ui_filename(self):
        return "filters.ui"

    def customize_device(self):
        filters = beamline.registry.findall(label="filters", allow_none=True)
        self.filters = sorted(
            filters, key=lambda dev: (dev.material.get(), dev.thickness.get())
        )

    def customize_ui(self):
        # Delete existing filter widgets
        for idx in reversed(range(self.filters_layout.count())):
            self.filters_layout.takeAt(idx).widget().deleteLater()
        # Add embedded displays for all the ion chambers
        self._filter_displays = []
        for idx, device in enumerate(self.filters):
            # Create the display object
            disp = PyDMEmbeddedDisplay(parent=self)
            disp.macros = json.dumps({"DEV": device.name})
            disp.filename = "filters_row.ui"
            # Add the Embedded Display to the Results Layout
            self.filters_layout.addWidget(disp)
            self._filter_displays.append(disp)



from queue import Queue

import numpy as np
import pandas as pd
from bluesky import plan_stubs as bps
from bluesky_adaptive.per_event import adaptive_plan, recommender_factory
from bluesky_adaptive.recommendations import NoRecommendation

from ..instrument import beamline

__all__ = ["GainRecommender", "auto_gain"]


class GainRecommender:
    """A recommendation engine for finding the best ion chamber gain*.

    Responds to ion chamber voltage as the dependent variable by
    changing the gain. If the voltage is above *volts_max* then the
    next gain* level will be higher, and if the voltage is below
    *volts_min* then the next gain* level will be lower. This engine
    will find all the values between *volts_min* and *volts_max*,
    assuming the output of the pre-amp changes monotonically with
    gain.

    """

    volts_min: float
    volts_max: float
    target_volts: float
    gain_min: int = 0
    gain_max: int = 27
    last_point: np.ndarray = None
    dfs: list = None
    big_step: int = 3

    def __init__(
        self, volts_min: float = 0.5, volts_max: float = 4.5, target_volts: float = 2.5
    ):
        self.volts_min = volts_min
        self.volts_max = volts_max
        self.target_volts = target_volts

    def tell(self, gains, volts):
        self.last_point = gains
        # Check if dataframes exist yet
        if self.dfs is None:
            self.dfs = [pd.DataFrame() for i in gains]
        # Add this measurement to the dataframes
        self.dfs = [
            pd.concat(
                [df, pd.DataFrame(data={"gain": [gain], "volts": [volt]})],
                ignore_index=True,
            )
            for gain, volt, df in zip(gains, volts, self.dfs)
        ]

    def tell_many(self, xs, ys):
        for x, y in zip(xs, ys):
            self.tell(x, y)

    def ask(self, n, tell_pending=True):
        """Figure out the next gain point based on the past ones we've measured."""
        if n != 1:
            raise NotImplementedError
        # Get the gains to try next
        next_point = [self.next_gain(df) for df in self.dfs]
        if np.array_equal(next_point, self.last_point):
            # We've already found the best point, so end the scan
            raise NoRecommendation
        return next_point

    def next_gain(self, df: pd.DataFrame):
        """Determine the next gain for this preamp based on past measurements.

        Parameters
        ==========
        df
          The past measurements for this preamp. Expected to have
          columns ["gain", "volts"].

        """
        # We're too low, so go up in gain
        if np.all(df.volts < self.volts_max):
            # Determine step size
            step = self.big_step if np.all(df.volts < self.volts_min) else 1
            # Determine next gain to use
            new_gain = df.gain.max() + step
            return np.min([new_gain, self.gain_max])
        # We're too high, so go down in gain
        if np.all(df.volts > self.volts_min):
            step = self.big_step if np.all(df.volts > self.volts_max) else 1
            new_gain = df.gain.min() - step
            return np.max([new_gain, self.gain_min])
        # Fill in any missing values through the correct gain
        values_in_range = df[(df.volts < self.volts_max) & (df.volts > self.volts_min)]
        needed_gains = np.arange(
            df[df.volts < self.volts_min].gain.max() + 1,
            df[df.volts > self.volts_max].gain.max(),
        )
        missing_gains = [
            gain for gain in needed_gains if gain not in values_in_range.gain
        ]
        if len(missing_gains) > 0:
            return max(missing_gains)
        # We have all the data we need, now decide on the best gain to use
        if len(values_in_range) > 0:
            good_vals = values_in_range
        else:
            good_vals = df
        best = good_vals.iloc[(good_vals.volts - self.target_volts).abs().argmin()]
        return best.gain


def auto_gain(
    ion_chambers="ion_chambers",
    volts_min: float = 0.5,
    volts_max: float = 4.5,
    prefer: str = "middle",
    max_count: int = 28,
    queue: Queue = None,
):
    """An adaptive Bluesky plan for optimizing ion chamber
    pre-amp gains.

    For each ion chamber, the plan will search for the range of gains
    within which the pre-amp output is between *volts_min* and
    *volts_max*, and select the gain that produces a voltage closest
    to the mid-point between *volts_min* and *volts_max*.

    Parameters
    ==========
    ion_chambers
      A sequence of detectors to scan. Can be devices, names, or Ophyd
      labels.
    volts_min
      The minimum acceptable range for each ion chamber's voltage.
    volts_max
      The maximum acceptable range for each ion chamber's voltage.
    prefer
      Whether to shoot for the "lower", "middle" (default), or "upper"
      portion of the voltage range.
    max_count
      The scan will end after *max_count* iterations even if an
      optimal gain has not been found for all pre-amps.
    queue
      [Testing] A Queue object for passing recommendations between the
      plan and the recommendation engine.

    """
    # Resolve the detector list into voltmeter AI's
    ion_chambers = beamline.registry.findall(ion_chambers)
    # Prepare the recommendation engine
    targets = {
        "lower": volts_min,
        "middle": (volts_min + volts_max) / 2,
        "upper": volts_max,
    }
    try:
        target = targets[prefer]
    except KeyError:
        raise ValueError(
            f"Invalid value for *prefer* {prefer}. Choices are 'lower', 'middle', or 'upper'."
        )
    recommender = GainRecommender(
        volts_min=volts_min, volts_max=volts_max, target_volts=target
    )
    preamp_gains = [det.preamp.gain_level for det in ion_chambers]
    ind_keys = [sig.name for sig in preamp_gains]
    voltmeters = [det.voltmeter_channel for det in ion_chambers]
    dep_keys = [voltmeter.final_value.name for voltmeter in voltmeters]
    rr, queue = recommender_factory(
        recommender,
        independent_keys=ind_keys,
        dependent_keys=dep_keys,
        max_count=max_count,
        queue=queue,
    )
    # Start from the current gain settings
    first_point = {}
    for det in ion_chambers:
        first_point[det.preamp.gain_level] = yield from bps.rd(
            det.preamp.gain_level, default_value=13
        )
    # Execute the plan
    yield from adaptive_plan(
        dets=voltmeters + preamp_gains,
        first_point=first_point,
        to_recommender=rr,
        from_recommender=queue,
    )


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest
from bluesky import RunEngine
from bluesky.callbacks.best_effort import BestEffortCallback
from ophyd import sim

from haven import align_motor

# from run_engine import RunEngineStub


@pytest.mark.skip(reason="Deprecated, use bluesky.plans.tune_centroid")
def test_align_motor(ffapp):
    # Set up simulated motors and detectors
    motor = sim.SynAxis(name="motor", labels={"motors"})
    detector = sim.SynGauss(
        name="detector",
        motor=motor,
        motor_field="motor",
        center=-3,
        Imax=1,
        sigma=20,
        labels={"detectors"},
    )
    # Prepare the callback to check results
    bec = BestEffortCallback()
    bec.disable_plots()
    bec.disable_table()
    # Prepare the plan
    plan = align_motor(
        detector=detector,
        motor=motor,
        distance=40,
        bec=bec,
        md={"plan_name": "test_plan"},
    )
    # Execute the plan
    RE = RunEngine(call_returns_result=True)
    result = RE(plan)
    # Check peak calculation results
    assert bec.peaks["cen"]["detector"] == pytest.approx(-3, rel=1e-3)
    assert motor.readback.get() == pytest.approx(-3, rel=1e-3)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------
